"""Utils for auto tasks"""
import copy
import warnings
import numpy as np

import autogluon.core as ag

from ..estimators.base_estimator import BaseEstimator
from ..estimators import SSDEstimator, FasterRCNNEstimator, YOLOv3Estimator, CenterNetEstimator
from ..estimators import ImageClassificationEstimator
from ..data.dataset import ObjectDetectionDataset

try:
    from ... import data as gdata
except ImportError:
    gdata = None


class ConfigDict(dict):
    """The view of a config dict where keys can be accessed like attribute, it also prevents
    naive modifications to the key-values.

    Parameters
    ----------
    config : dict
        The configuration dict.

    Attributes
    ----------
    __dict__ : type
        The internal config as a `__dict__`.

    """
    MARKER = object()
    def __init__(self, value=None):
        super(ConfigDict, self).__init__()
        self.__dict__['_freeze'] = False
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError('expected dict, given {}'.format(type(value)))
        self.freeze()

    def freeze(self):
        self.__dict__['_freeze'] = True

    def is_frozen(self):
        return self.__dict__['_freeze']

    def unfreeze(self):
        self.__dict__['_freeze'] = False

    def __setitem__(self, key, value):
        if self.__dict__.get('_freeze', False):
            msg = ('You are trying to modify the config to "{}={}" after initialization, '
                   ' this may result in unpredictable behaviour'.format(key, value))
            warnings.warn(msg)
        if isinstance(value, dict) and not isinstance(value, ConfigDict):
            value = ConfigDict(value)
        super(ConfigDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        found = self.get(key, ConfigDict.MARKER)
        if found is ConfigDict.MARKER:
            if self.__dict__['_freeze']:
                raise KeyError(key)
            found = ConfigDict()
            super(ConfigDict, self).__setitem__(key, found)
        if isinstance(found, ConfigDict):
            found.__dict__['_freeze'] = self.__dict__['_freeze']
        return found

    def __setstate__(self, state):
        vars(self).update(state)

    def __getstate__(self):
        return vars(self)

    __setattr__, __getattr__ = __setitem__, __getitem__

def auto_suggest(config, estimator, logger):
    """
    Automatically suggest some hyperparameters based on the dataset statistics.
    """
    # specify estimator search space
    if estimator is None:
        estimator_init = [SSDEstimator, YOLOv3Estimator, FasterRCNNEstimator, CenterNetEstimator]
        config['estimator'] = ag.Categorical(*estimator_init)
    elif isinstance(estimator, str):
        named_estimators = {
            'ssd': SSDEstimator,
            'faster_rcnn': FasterRCNNEstimator,
            'yolo3': YOLOv3Estimator,
            'center_net': CenterNetEstimator,
            'img_cls': ImageClassificationEstimator
        }
        if estimator.lower() in named_estimators:
            estimator = [named_estimators[estimator.lower()]]
        else:
            available_ests = named_estimators.keys()
            raise ValueError(f'Unknown estimator name: {estimator}, options: {available_ests}')
    elif isinstance(estimator, (tuple, list)):
        pass
    else:
        if isinstance(estimator, ag.Space):
            estimator = estimator.data
        elif isinstance(estimator, str):
            estimator = [estimator]
        for i, e in enumerate(estimator):
            if e == 'ssd':
                estimator[i] = SSDEstimator
            elif e == 'yolo3':
                estimator[i] = YOLOv3Estimator
            elif e == 'faster_rcnn':
                estimator[i] = FasterRCNNEstimator
            elif e == 'center_net':
                estimator[i] = CenterNetEstimator
        if not estimator:
            raise ValueError('Unable to determine the estimator for fit function.')
        if len(estimator) == 1:
            config['estimator'] = estimator[0]
        else:
            config['estimator'] = ag.Categorical(*estimator)

    # get dataset statistics
    # user needs to define a Dataset object "train_dataset" when using custom dataset
    train_dataset = config.get('train_dataset', None)
    try:
        if train_dataset is None:
            dataset_name = config.get('dataset', 'voc')
            dataset_root = config.get('dataset_root', '~/.mxnet/datasets/')
            if gdata is not None and dataset_name == 'voc':
                train_dataset = gdata.VOCDetection(splits=[(2007, 'trainval'), (2012, 'trainval')])
            elif gdata is not None and dataset_name == 'voc_tiny':
                train_dataset = gdata.CustomVOCDetectionBase(classes=('motorbike',),
                                                             root=dataset_root + 'tiny_motorbike',
                                                             splits=[('', 'trainval')])
            elif gdata is not None and dataset_name == 'coco':
                train_dataset = gdata.COCODetection(splits=['instances_train2017'])
            else:
                if gdata is None:
                    logger.warning('Unable to import mxnet so voc and coco formats are currently not loaded.')
        elif isinstance(train_dataset, ObjectDetectionDataset):
            train_dataset = train_dataset.to_mxnet()
        else:
            logger.info('Unknown dataset, quit auto suggestion...')
            return
    # pylint: disable=broad-except
    except Exception as e:
        logger.info(f'Unexpected error: {e}, quit auto suggestion...')
        return

    # choose 100 examples to calculate average statistics
    num_examples = 100
    image_size_list = []
    num_objects_list = []
    bbox_size_list = []
    bbox_rel_size_list = []

    for i in range(num_examples):
        train_image, train_label = train_dataset[i]

        image_height = train_image.shape[0]
        image_width = train_image.shape[1]
        image_size = image_height * image_width
        image_size_list.append(image_size)

        bounding_boxes = train_label[:, :4]
        num_objects = bounding_boxes.shape[0]
        bbox_height = bounding_boxes[:, 3] - bounding_boxes[:, 1]
        bbox_width = bounding_boxes[:, 2] - bounding_boxes[:, 0]
        bbox_size = bbox_height * bbox_width
        bbox_rel_size = bbox_size / image_size
        num_objects_list.append(num_objects)
        bbox_size_list.append(np.mean(bbox_size))
        bbox_rel_size_list.append(np.mean(bbox_rel_size))

    num_images = len(train_dataset)
    image_size = np.mean(image_size_list)
    try:
        num_classes = len(train_dataset.CLASSES)
    except AttributeError:
        num_classes = len(train_dataset.classes)
    num_objects = np.mean(num_objects_list)
    bbox_size = np.mean(bbox_size_list)
    bbox_rel_size = np.mean(bbox_rel_size_list)

    logger.info("[Printing dataset statistics]...")
    logger.info("number of training images: %d", num_images)
    logger.info("average image size: %.2f", image_size)
    logger.info("number of total object classes: %d", num_classes)
    logger.info("average number of objects in an image: %.2f", num_objects)
    logger.info("average bounding box size: %.2f", bbox_size)
    logger.info("average bounding box relative size: %.2f", bbox_rel_size)

    # specify 3 parts of config: data preprocessing, model selection, training settings
    if bbox_rel_size < 0.2 or num_objects > 5:
        suggested_estimator = [FasterRCNNEstimator]
    else:
        suggested_estimator = [SSDEstimator, YOLOv3Estimator, CenterNetEstimator]

    # specify estimator search space based on suggestion
    if estimator is None:
        estimator = suggested_estimator
        config['estimator'] = ag.Categorical(*estimator)

def get_recursively(search_dict, field):
    """
    Takes a dict with nested dicts,
    and searches all dicts for a key of the field
    provided.
    """
    fields_found = []

    for key, value in search_dict.items():

        if key == field:
            fields_found.append(value)

        elif isinstance(value, dict):
            results = get_recursively(value, field)
            for result in results:
                fields_found.append(result)

    return fields_found

def config_to_nested(config):
    """Convert config to nested version"""
    estimator = config.get('estimator', None)
    transfer = config.get('transfer', None)
    # choose hyperparameters based on pretrained model in transfer learning
    if transfer:
        # choose estimator
        if transfer.startswith('ssd'):
            estimator = SSDEstimator
        elif transfer.startswith('yolo3'):
            estimator = YOLOv3Estimator
        elif transfer.startswith('faster_rcnn'):
            estimator = FasterRCNNEstimator
        elif transfer.startswith('center_net'):
            estimator = CenterNetEstimator
        else:
            estimator = ImageClassificationEstimator
        # choose base network
        transfer_list = transfer.split('_')
        if transfer_list[0] == 'ssd':
            transfer_list.pop(0)
            config['data_shape'] = int(transfer_list.pop(0))
            transfer_list.pop(-1)
        elif transfer_list[0] == 'yolo3':
            transfer_list.pop(0)
            transfer_list.pop(-1)
        else:
            transfer_list.pop(0)
            transfer_list.pop(0)
            transfer_list.pop(-1)
        config['base_network'] = '_'.join(transfer_list)
        if config['base_network'].startswith('mobilenet'):
            config['base_network'].replace('_', '.')
    elif isinstance(estimator, str):
        if estimator == 'ssd':
            estimator = SSDEstimator
        elif estimator == 'yolo3':
            estimator = YOLOv3Estimator
        elif estimator == 'faster_rcnn':
            estimator = FasterRCNNEstimator
        elif estimator == 'center_net':
            estimator = CenterNetEstimator
        elif estimator == 'img_cls':
            estimator = ImageClassificationEstimator
        else:
            raise ValueError(f'Unknown estimator: {estimator}')
    else:
        assert issubclass(estimator, BaseEstimator)

    cfg_map = estimator._default_cfg.asdict()

    def _recursive_update(config, key, value, auto_strs, auto_ints):
        for k, v in config.items():
            if k in auto_strs:
                config[k] = 'auto'
            if k in auto_ints:
                config[k] = -1
            if key == k:
                config[key] = value
            elif isinstance(v, dict):
                _recursive_update(v, key, value, auto_strs, auto_ints)

    if 'use_rec' in config:
        auto_strs = ['data_dir']
        auto_ints = []
    else:
        auto_strs = ['data_dir', 'rec_train', 'rec_train_idx', 'rec_val', 'rec_val_idx',
                     'dataset', 'dataset_root']
        auto_ints = ['num_training_samples']
    for k, v in config.items():
        _recursive_update(cfg_map, k, v, auto_strs, auto_ints)
    cfg_map['estimator'] = estimator
    return cfg_map

def config_to_nested_v0(config):
    """Convert config to nested version"""
    if 'meta_arch' not in config:
        if config['estimator'] == SSDEstimator:
            config['meta_arch'] = 'ssd'
        elif config['estimator'] == FasterRCNNEstimator:
            config['meta_arch'] = 'faster_rcnn'
        elif config['estimator'] == YOLOv3Estimator:
            config['meta_arch'] = 'yolo3'
        elif config['estimator'] == CenterNetEstimator:
            config['meta_arch'] = 'center_net'
        elif config['estimator'] == ImageClassificationEstimator:
            config['meta_arch'] = 'img_cls'
        else:
            config['meta_arch'] = None
    else:
        pass

    if config['meta_arch'] == 'ssd':
        config_mapping = {
            'ssd': ['backbone', 'data_shape', 'filters', 'sizes', 'ratios', 'steps', 'syncbn',
                    'amp', 'custom_model'],
            'train': ['batch_size', 'start_epoch', 'epochs', 'lr', 'lr_decay', 'lr_decay_epoch',
                      'momentum', 'wd', 'log_interval', 'seed', 'dali'],
            'validation': ['val_interval']
        }
    elif config['meta_arch'] == 'faster_rcnn':
        config_mapping = {
            'faster_rcnn': ['backbone', 'nms_thresh', 'nms_topk', 'roi_mode', 'roi_size', 'strides', 'clip',
                            'anchor_base_size', 'anchor_aspect_ratio', 'anchor_scales', 'anchor_alloc_size',
                            'rpn_channel', 'rpn_nms_thresh', 'max_num_gt', 'norm_layer', 'use_fpn', 'num_fpn_filters',
                            'num_box_head_conv', 'num_box_head_conv_filters', 'num_box_head_dense_filters',
                            'image_short', 'image_max_size', 'custom_model', 'amp', 'static_alloc',
                            'disable_hybridization'],
            'train': ['pretrained_base', 'batch_size', 'start_epoch', 'epochs', 'lr', 'lr_decay',
                      'lr_decay_epoch', 'lr_mode', 'lr_warmup', 'lr_warmup_factor', 'momentum', 'wd',
                      'rpn_train_pre_nms', 'rpn_train_post_nms', 'rpn_smoothl1_rho', 'rpn_min_size',
                      'rcnn_num_samples', 'rcnn_pos_iou_thresh', 'rcnn_pos_ratio', 'rcnn_smoothl1_rho',
                      'log_interval', 'seed', 'verbose', 'mixup', 'no_mixup_epochs', 'executor_threads'],
            'validation': ['rpn_test_pre_nms', 'rpn_test_post_nms', 'val_interval']
        }
    elif config['meta_arch'] == 'yolo3':
        config_mapping = {
            'yolo3': ['backbone', 'filters', 'anchors', 'strides', 'data_shape', 'syncbn', 'no_random_shape',
                      'amp', 'custom_model'],
            'train': ['batch_size', 'epochs', 'start_epoch', 'lr', 'lr_mode', 'lr_decay', 'lr_decay_period',
                      'lr_decay_epoch', 'warmup_lr', 'warmup_epochs', 'momentum', 'wd', 'log_interval',
                      'seed', 'num_samples', 'no_wd', 'mixup', 'no_mixup_epochs', 'label_smooth'],
            'validation': ['val_interval']
        }
    elif config['meta_arch'] == 'center_net':
        config_mapping = {
            'center_net': ['base_network', 'heads', 'scale', 'topk', 'root', 'wh_weight', 'center_reg_weight',
                           'data_shape'],
            'train': ['gpus', 'pretrained_base', 'batch_size', 'epochs', 'lr', 'lr_decay', 'lr_decay_epoch',
                      'lr_mode', 'warmup_lr', 'warmup_epochs', 'num_workers', 'resume',
                      'start_epoch', 'momentum', 'wd', 'save_interval', 'log_interval'],
            'validation': ['flip_test', 'nms_thresh', 'nms_topk', 'post_nms', 'num_workers',
                           'batch_size', 'interval']
        }
    elif config['meta_arch'] == 'img_cls':
        config_mapping = {
            'img_cls': ['model', 'use_pretrained', 'use_gn', 'batch_norm', 'use_se', 'last_gamma'],
            'train': ['gpus', 'num_workers', 'batch_size', 'epochs', 'start_epoch', 'lr', 'lr_mode',
                      'lr_decay', 'lr_decay_period',
                      'lr_decay_epoch', 'warmup_lr', 'warmup_epochs', 'momentum', 'wd', 'log_interval',
                      'seed', 'num_samples', 'no_wd', 'mixup', 'no_mixup_epochs', 'label_smooth'],
            'validation': []
        }
    else:
        raise NotImplementedError('%s is not implemented.' % config['meta_arch'])

    nested_config = {}

    for k, v in config.items():
        if k in config_mapping[config['meta_arch']]:
            if config['meta_arch'] not in nested_config:
                nested_config[config['meta_arch']] = {}
            nested_config[config['meta_arch']].update({k: v})
        elif k in config_mapping['train']:
            if 'train' not in nested_config:
                nested_config['train'] = {}
            nested_config['train'].update({k: v})
        elif k in config_mapping['validation']:
            if 'validation' not in nested_config:
                nested_config['validation'] = {}
            nested_config['validation'].update({k: v})
        else:
            nested_config.update({k: v})

    return nested_config

def recursive_update(total_config, config):
    """update config recursively"""
    for k, v in config.items():
        if isinstance(v, dict):
            if k not in total_config:
                total_config[k] = {}
            recursive_update(total_config[k], v)
        else:
            total_config[k] = v

def config_to_space(config):
    """Convert config to ag.space"""
    space = ag.Dict()
    for k, v in config.items():
        if isinstance(v, dict):
            if k not in space:
                space[k] = ag.Dict()
            space[k] = config_to_space(v)
        else:
            space[k] = v
    return space

def auto_args(config, estimators):
    """
    Merge user defined config to estimator default config, and convert to search space

    Parameters
    ----------
    config: <class 'dict'>

    Returns
    -------
    ag_space: <class 'autogluon.core.space.Dict'>
    """
    total_config = {}

    # estimator default config
    if not isinstance(estimators, (tuple, list)):
        estimators = [estimators]
    for estimator in estimators:
        assert issubclass(estimator, BaseEstimator), estimator
        default_config = copy.deepcopy(estimator._default_config)  # <class 'dict'>
        recursive_update(total_config, default_config)

    # user defined config
    nested_config = config_to_nested(config)
    recursive_update(total_config, nested_config)

    # convert to search space
    ag_space = config_to_space(total_config)

    return ag_space



"""
Utility functions for COOT model
"""
import ctypes
import datetime
import logging
import multiprocessing as mp
import os
from pathlib import Path
import random
import sys
from typing import Tuple, Dict

import numpy as np
import torch
import torch.backends.cudnn as cudnn
from torch import cuda
import torch.nn.functional as F

EVALKEYS = ["r1", "r5", "r10", "r50", "medr", "meanr", "sum"]
EVALHEADER = "Retriev | R@1   | R@5   | R@10  | R@50  | MeanR |  MedR |    Sum"


def create_dataloader_path(data_root,
                           dataset_name='youcook2',
                           text_feature_name='default',
                           video_feature_name='100m'):
    """create the path to meta file and features

    Args:
        data_root ([PATH]): [Path to the data folder]

    Returns:
        [Dict]: [path to meta data and video/language features]
    """

    meta_data_path = Path(
        os.path.join(data_root, "meta_{}.json".format(video_feature_name)))
    video_feat_path = Path(
        os.path.join(data_root, "video_feat_{}.h5".format(video_feature_name)))
    language_feat_path = Path(
        os.path.join(data_root, "text_{}.h5".format(text_feature_name)))
    meta_text_len_path = Path(
        os.path.join(data_root, "text_lens_{}.json".format(text_feature_name)))

    return {
        "dataset_name": dataset_name,
        "meta_data": meta_data_path,
        "video_feats": video_feat_path,
        "language_feats": language_feat_path,
        "meta_text_len": meta_text_len_path,
    }


def get_csv_header_keys(compute_clip_retrieval):
    """ get CSV header keys"""

    metric_keys = ["ep", "time"]
    prefixes = ["v", "p"]
    if compute_clip_retrieval:
        prefixes += ["c", "s"]
    for prefix in prefixes:
        for key in EVALKEYS:
            metric_keys.append(f"{prefix}-{key}")
    return metric_keys


def expand_segment(num_frames, num_target_frames, start_frame, stop_frame):
    """ expand the segment"""
    num_frames_seg = stop_frame - start_frame + 1
    changes = False
    num_target_frames = min(num_target_frames, num_frames)
    if num_frames_seg < num_target_frames:
        while True:
            if start_frame > 0:
                start_frame -= 1
                num_frames_seg += 1
                changes = True
            if num_frames_seg == num_target_frames:
                break
            if stop_frame < num_frames - 1:
                stop_frame += 1
                num_frames_seg += 1
                changes = True
            if num_frames_seg == num_target_frames:
                break
    return start_frame, stop_frame, changes


def set_seed(seed: int) -> None:
    """ set seed"""
    torch.manual_seed(seed)
    cuda.manual_seed(seed)
    cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    cudnn.benchmark = False
    cudnn.deterministic = True


def make_shared_array(np_array: np.ndarray) -> mp.Array:
    """ shared array"""
    flat_shape = int(np.prod(np_array.shape))
    shared_array_base = mp.Array(ctypes.c_float, flat_shape)
    shared_array = np.ctypeslib.as_array(shared_array_base.get_obj())
    shared_array = shared_array.reshape(np_array.shape)
    shared_array[:] = np_array[:]
    return shared_array


def compute_indices(num_frames_orig: int, num_frames_target: int,
                    is_train: bool):
    """ compute indices """
    def round_half_down(array: np.ndarray) -> np.ndarray:
        return np.ceil(array - 0.5)

    if is_train:
        # random sampling during training
        start_points = np.linspace(0,
                                   num_frames_orig,
                                   num_frames_target,
                                   endpoint=False)
        start_points = round_half_down(start_points).astype(int)
        offsets = start_points[1:] - start_points[:-1]
        np.random.shuffle(offsets)
        last_offset = num_frames_orig - np.sum(offsets)
        offsets = np.concatenate([offsets, np.array([last_offset])])
        new_start_points = np.cumsum(offsets) - offsets[0]
        offsets = np.roll(offsets, -1)
        random_offsets = offsets * np.random.rand(num_frames_target)
        indices = new_start_points + random_offsets
        indices = np.floor(indices).astype(int)
        return indices
    # center sampling during validation
    start_points = np.linspace(0,
                               num_frames_orig,
                               num_frames_target,
                               endpoint=False)
    offset = num_frames_orig / num_frames_target / 2
    indices = start_points + offset
    indices = np.floor(indices).astype(int)
    return indices


def truncated_normal_fill(shape: Tuple[int],
                          mean: float = 0,
                          std: float = 1,
                          limit: float = 2) -> torch.Tensor:
    """ truncate normal """

    num_examples = 8
    tmp = torch.empty(shape + (num_examples, )).normal_()
    valid = (tmp < limit) & (tmp > -limit)
    _, ind = valid.max(-1, keepdim=True)
    return tmp.gather(-1, ind).squeeze(-1).mul_(std).add_(mean)


def retrieval_results_to_str(results: Dict[str, float], name: str):
    """ retrieval results string """
    return ("{:7s} | {:.3f} | {:.3f} | {:.3f} | {:.3f} | {:5.1f} | "
            "{:5.1f} | {:6.3f}").format(name, *[results[a] for a in EVALKEYS])


def compute_retr_vid_to_par(video_feat, cap_feat):
    """ compute similarity scores video to paragraph """
    similarity_scores = np.dot(video_feat, cap_feat.T)
    return compute_retrieval_metrics(similarity_scores)


def compute_retr_par_to_vid(video_feat, cap_feat):
    """ compute similarity scores paragraph to video """
    similarity_scores = np.dot(cap_feat, video_feat.T)
    return compute_retrieval_metrics(similarity_scores)


def compute_retrieval_metrics(dot_product):
    """ Compute the retrieval performance

    Args:
        dot_product (similarity of embeddings X1 and X2): dot_product(X1, X2)

    Returns:
        Retrieval evaluation metrics such as R1, R5 and so on.
    """

    sort_similarity = np.sort(-dot_product, axis=1)
    diag_similarity = np.diag(-dot_product)
    diag_similarity = diag_similarity[:, np.newaxis]
    ranks = sort_similarity - diag_similarity
    ranks = np.where(ranks == 0)
    ranks = ranks[1]

    report_dict = dict()
    report_dict['r1'] = float(np.sum(ranks == 0)) / len(ranks)
    report_dict['r5'] = float(np.sum(ranks < 5)) / len(ranks)
    report_dict['r10'] = float(np.sum(ranks < 10)) / len(ranks)
    report_dict['r50'] = float(np.sum(ranks < 50)) / len(ranks)
    report_dict['medr'] = np.median(ranks) + 1
    report_dict['meanr'] = ranks.mean()
    report_dict[
        'sum'] = report_dict['r1'] + report_dict['r5'] + report_dict['r50']

    return report_dict, ranks


def compare_metrics(comparison, best):
    """ compare metrics """

    if best is None:
        return True
    threshold = 1e-4
    rel_epsilon = threshold + 1.
    return comparison > best * rel_epsilon


def get_logging_formatter():
    """ logging formatter """
    return logging.Formatter("%(asctime)s %(levelname)s %(message)s",
                             datefmt="%m%d %H%M%S")


def get_timestamp_for_filename():
    """ timestamp"""

    time_split = str(datetime.datetime.now()).split(".")[0].replace(" ", "_")
    time_split = time_split.replace(":", "_").replace("-", "_")
    return time_split


def get_logger_without_file(name, log_level="INFO") -> logging.Logger:
    """ gett basic logger"""

    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    strm_hdlr = logging.StreamHandler(sys.stdout)
    strm_hdlr.setFormatter(get_logging_formatter())
    logger.addHandler(strm_hdlr)
    return logger


def get_logger(logdir,
               name,
               filename="run",
               log_level="INFO",
               log_file=True) -> logging.Logger:
    """Get logger

    Returns:
        logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    formatter = get_logging_formatter()
    if log_file:
        file_path = Path(logdir) / "{}_{}.log".format(
            filename,
            str(datetime.datetime.now()).split(".")[0].replace(
                " ", "_").replace(":", "_").replace("-", "_"))
        file_hdlr = logging.FileHandler(str(file_path))
        file_hdlr.setFormatter(formatter)
        logger.addHandler(file_hdlr)
    strm_hdlr = logging.StreamHandler(sys.stdout)
    strm_hdlr.setFormatter(formatter)
    logger.addHandler(strm_hdlr)
    logger.propagate = False
    return logger


def close_logger(logger: logging.Logger):
    """ close logger """

    log_handle_list = list(logger.handlers)
    for i in log_handle_list:
        logger.removeHandler(i)
        i.flush()
        i.close()


def unpack_data(data_dict, use_cuda):
    """unpack data
    """
    def to_device(x):
        if use_cuda and isinstance(x, torch.Tensor):
            return x.cuda(non_blocking=True)
        return x

    return [
        to_device(data_dict[a])
        for a in ("vid_id", "vid_frames", "vid_frames_mask", "vid_frames_len",
                  "par_cap_vectors", "par_cap_mask", "par_cap_len", "clip_num",
                  "clip_frames", "clip_frames_len", "clip_frames_mask",
                  "sent_num", "sent_cap_vectors", "sent_cap_mask",
                  "sent_cap_len")
    ]


def compute_constrastive_loss(config, contrastive_loss, vid_emb, par_emb,
                              clip_emb, sent_emb, vid_context, par_context):
    """Normalize embeddings and calculate alignment loss in different levels:
     Video-paragraph, clip-sentence, global context

    Args:
        contrastive_loss (loss function): MaxMargingRanking loss
        vid_emb (tensor): video embeddings with shape batch*dim
        par_emb (tensor): paragraph embeddings with shape batch*dim
        clip_emb (tensor): clip embeddings
        sent_emb (tensor): sentence embeddings
        vid_context (tensor): video global context
        par_context (tensor): paragraph global context

    Returns:
        total loss
    """
    vid_context_norm = F.normalize(vid_context)
    clip_emb_norm = F.normalize(clip_emb)
    vid_emb_norm = F.normalize(vid_emb)
    par_context_norm = F.normalize(par_context)
    sent_emb_norm = F.normalize(sent_emb)
    par_emb_norm = F.normalize(par_emb)

    loss = contrastive_loss(vid_emb_norm, par_emb_norm)
    loss += config.CONFIG.TRAIN.LOSS_CONTRASTIVE_CLIP_W * contrastive_loss(
        clip_emb_norm, sent_emb_norm)
    loss += contrastive_loss(vid_context_norm, par_context_norm)
    loss += (contrastive_loss(vid_emb_norm, vid_emb_norm) +
             contrastive_loss(par_emb_norm, par_emb_norm)) / 2
    loss += (contrastive_loss(clip_emb_norm, clip_emb_norm) +
             contrastive_loss(sent_emb_norm, sent_emb_norm)) / 2
    return loss


def compute_cmc_loss(cyc_consistency_loss, loss_weight, clip_emb_reshape,
                     clip_emb_mask, clip_emb_lens, sent_emb_reshape,
                     sent_emb_mask, sent_emb_lens):
    """Calculate the total cycle consistency loss between video clips and paragraph sentences

    Args:
        cyc_consistency_loss (loss function): cycle consistency loss function
        loss_weight (float): weight of loss
        clip_emb_reshape
        clip_emb_mask
        clip_emb_lens
        sent_emb_reshape
        sent_emb_mask
        sent_emb_lens

    Returns:
        total loss
    """
    clip_clip_loss, sent_sent_loss = cyc_consistency_loss(
        clip_emb_reshape, clip_emb_mask, clip_emb_lens, sent_emb_reshape,
        sent_emb_mask, sent_emb_lens)
    loss = loss_weight * (clip_clip_loss + sent_sent_loss)
    return loss



"""Market 1501 Person Re-Identification Dataset."""
from mxnet.gluon.data import dataset
from mxnet import image

__all__ = ['ImageTxtDataset']


class ImageTxtDataset(dataset.Dataset):
    """Load the Market 1501 dataset.

    Parameters
    ----------
    items : list
        List for image names and labels.
    flag : int, default 1
        Whether to load the color image or gray image.
    transform : function, default None
        A function that takes data and label and transforms them.
    """
    def __init__(self, items, flag=1, transform=None):
        self._flag = flag
        self._transform = transform
        self.items = items

    def __getitem__(self, idx):
        fpath = self.items[idx][0]
        img = image.imread(fpath, self._flag)
        label = self.items[idx][1]
        if self._transform is not None:
            img = self._transform(img)
        return img, label

    def __len__(self):
        return len(self.items)



"""Avoid filename collision"""
import logging
import os
import hashlib
import re
from pathlib import Path
import tqdm

from ..dataset import GluonCVMotionDataset
from ..utils.serialization_utils import save_json


_log = logging.getLogger()
_log.setLevel(logging.DEBUG)

PATH_MAXLEN = 255
non_word_pattern = re.compile(r'[^A-Za-z0-9_\- ]+')


def hash_path(path, algorithm="blake2b", suffix_len=None, num_orig_chars=None, constant_len=False):
    # pylint: disable=missing-function-docstring
    path = Path(path)
    if suffix_len is None:
        suffix_len = len(path.suffix)
    stem = str(path.stem)
    replaced_stem = stem.replace(' ', '_')
    replaced_stem = replaced_stem.replace('-', '_')
    filtered_stem = non_word_pattern.sub('', replaced_stem)
    if len(filtered_stem) == len(stem):
        return replaced_stem

    path = str(path)
    if algorithm == "blake2b":
        # Using blake2b by default since it is fast and as good as sha-3: https://blake2.net/
        hashstr = hashlib.blake2b(path.encode(), digest_size=16).hexdigest()
    elif algorithm == "md5":
        hashstr = hashlib.md5(path.encode()).hexdigest()
    else:
        raise ValueError("Unsupported algorithm {}".format(algorithm))

    # 1 for underscore between
    max_orig_chars = PATH_MAXLEN - (len(hashstr) + 1) - suffix_len

    orig_take_chars = max_orig_chars if num_orig_chars is None \
                                     else min(num_orig_chars, max_orig_chars)
    if orig_take_chars > 0:
        trunc_stem = filtered_stem[:orig_take_chars]
        if num_orig_chars and constant_len:
            trunc_stem = trunc_stem.ljust(orig_take_chars, '_')
        new_stem = "{}_{}".format(trunc_stem, hashstr)
    else:
        new_stem = hashstr

    return new_stem


def main(anno_path, cache_name="uuid_filenames", ref_cache=None,
         ref_ext=".mp4", flatten=False, no_ext_hash=False):
    # pylint: disable=line-too-long， missing-function-docstring
    dataset = GluonCVMotionDataset(anno_path)

    id_mapping = {}
    cache_dir = Path(dataset.cache_root_path) / cache_name

    new_path_set = set()
    samples = dataset.samples
    for sample_id, sample in tqdm.tqdm(samples, mininterval=1):
        if ref_cache:
            ref_filepath = sample.get_cache_file(ref_cache, ref_ext)
        else:
            ref_filepath = sample.data_path

        orig_cache_filepath = Path(sample.get_cache_file(cache_name, Path(ref_filepath).suffix))
        orig_suffix = orig_cache_filepath.suffix
        orig_cache_rel_filepath = orig_cache_filepath.relative_to(cache_dir)
        if no_ext_hash:
            orig_cache_rel_filepath = orig_cache_rel_filepath.with_suffix("")
        orig_filename = orig_cache_rel_filepath.name

        if flatten:
            hashed_filepath = hash_path(orig_cache_rel_filepath, suffix_len=len(orig_suffix))
            uuid_rel_filepath = hashed_filepath + orig_suffix
        else:
            hashed_filename = hash_path(orig_filename, suffix_len=len(orig_suffix))
            uuid_rel_filepath = orig_cache_rel_filepath.with_name(hashed_filename + orig_suffix)
        uuid_filepath = cache_dir / uuid_rel_filepath
        if uuid_rel_filepath in new_path_set:
            existing_id = [id for id, p in id_mapping.items() if p == str(uuid_rel_filepath)]
            _log.error("Unexpected filename collision with: {} , \
                       existing id: {} , new sample id: {}".format(uuid_rel_filepath, existing_id, sample_id))
            continue
        new_path_set.add(uuid_rel_filepath)
        id_mapping[sample_id] = str(uuid_rel_filepath)
        os.makedirs(uuid_filepath.parent, exist_ok=True)
        if not os.path.exists(uuid_filepath):
            os.symlink(ref_filepath, uuid_filepath)

    save_json(id_mapping, cache_dir / "id_mapping.json")


if __name__ == "__main__":
    import fire
    fire.Fire(main)



# pylint: disable=missing-function-docstring, unused-argument
"""
C2D video action recognition models based on ResNet
"""
import torch
import torch.nn as nn
import torchvision


__all__ = ['ActionRecResNetV1b', 'resnet18_v1b_kinetics400', 'resnet34_v1b_kinetics400',
           'resnet50_v1b_kinetics400', 'resnet101_v1b_kinetics400', 'resnet152_v1b_kinetics400',
           'resnet50_v1b_sthsthv2', 'resnet50_v1b_custom']


class ActionRecResNetV1b(nn.Module):
    r"""ResNet models for video action recognition
    Deep Residual Learning for Image Recognition, CVPR 2016
    https://arxiv.org/abs/1512.03385

    Parameters
    ----------
    depth : int, default is 50.
        Depth of ResNet, from {18, 34, 50, 101, 152}.
    num_classes : int
        Number of classes in the training dataset.
    pretrained_base : bool, default is True.
        Load pretrained base network (backbone), the extra layers are randomized.
    feat_ext : bool, default is False.
        Whether to extract features from backbone network or perform a standard network forward.
    partial_bn : bool, default is False.
        Freeze all batch normalization layers during training except the first one.
    dropout_ratio : float, default is 0.5.
        The dropout rate of a dropout layer.
        The larger the value, the more strength to prevent overfitting.
    init_std : float, default is 0.01.
        Standard deviation value when initialize a fully connected layer.
    num_segment : int, default is 1.
        Number of segments used to evenly divide a video.
    num_crop : int, default is 1.
        Number of crops used during training.

    Input : a single video frame or N images from N segments when num_segment > 1
    Output : a single predicted action label
    """
    def __init__(self, depth, num_classes, pretrained_base=True,
                 dropout_ratio=0.5, init_std=0.01,
                 num_segment=1, num_crop=1, feat_ext=False,
                 partial_bn=False, **kwargs):
        super(ActionRecResNetV1b, self).__init__()

        self.depth = depth
        self.num_classes = num_classes
        self.pretrained_base = pretrained_base
        self.feat_ext = feat_ext

        if self.depth == 18:
            C2D = torchvision.models.resnet18(pretrained=self.pretrained_base, progress=True)
            self.expansion = 1
        elif self.depth == 34:
            C2D = torchvision.models.resnet34(pretrained=self.pretrained_base, progress=True)
            self.expansion = 1
        elif self.depth == 50:
            C2D = torchvision.models.resnet50(pretrained=self.pretrained_base, progress=True)
            self.expansion = 4
        elif self.depth == 101:
            C2D = torchvision.models.resnet101(pretrained=self.pretrained_base, progress=True)
            self.expansion = 4
        elif self.depth == 152:
            C2D = torchvision.models.resnet152(pretrained=self.pretrained_base, progress=True)
            self.expansion = 4
        else:
            raise RuntimeError("We do not support ResNet with depth %d." % (self.depth))

        self.dropout_ratio = dropout_ratio
        self.init_std = init_std
        self.num_segment = num_segment
        self.num_crop = num_crop
        self.feat_dim = 512 * self.expansion

        self.conv1 = C2D.conv1
        self.bn1 = C2D.bn1
        self.relu = C2D.relu
        self.maxpool = C2D.maxpool
        self.layer1 = C2D.layer1
        self.layer2 = C2D.layer2
        self.layer3 = C2D.layer3
        self.layer4 = C2D.layer4
        self.avgpool = C2D.avgpool

        self.drop = nn.Dropout(self.dropout_ratio)
        self.fc = nn.Linear(in_features=self.feat_dim, out_features=num_classes)
        nn.init.normal_(self.fc.weight, 0, self.init_std)
        nn.init.constant_(self.fc.bias, 0)

    def forward(self, x):
        bs, ch, tm, h, w = x.shape
        x = x.permute(0, 2, 1, 3, 4)
        x = x.contiguous().view(bs * tm, ch, h, w)

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = self.drop(x)
        x = torch.flatten(x, 1)

        # segmental consensus
        x = x.view(bs, tm, self.feat_dim)
        x = torch.mean(x, dim=1)

        if self.feat_ext:
            return x

        x = self.fc(x)
        return x


def resnet18_v1b_kinetics400(cfg):
    model = ActionRecResNetV1b(depth=18,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet18_v1b_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet34_v1b_kinetics400(cfg):
    model = ActionRecResNetV1b(depth=34,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet34_v1b_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet50_v1b_kinetics400(cfg):
    model = ActionRecResNetV1b(depth=50,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet50_v1b_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet101_v1b_kinetics400(cfg):
    model = ActionRecResNetV1b(depth=101,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet101_v1b_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet152_v1b_kinetics400(cfg):
    model = ActionRecResNetV1b(depth=152,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet152_v1b_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet50_v1b_sthsthv2(cfg):
    model = ActionRecResNetV1b(depth=50,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('resnet50_v1b_sthsthv2',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def resnet50_v1b_custom(cfg):
    model = ActionRecResNetV1b(depth=50,
                               num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                               pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                               feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                               partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                               num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                               num_crop=cfg.CONFIG.DATA.NUM_CROP,
                               dropout_ratio=0.5,
                               init_std=0.01)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('resnet50_v1b_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias'}
        print("=> Initialized from a ResNet50 model pretrained on Kinetcis400 dataset")
    return model



# -*- coding: utf-8 -*-
# adapted from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/transforms/augmentation.py
import sys
import inspect
import random
import numpy as np
import pprint
from abc import ABCMeta, abstractmethod
from typing import List, Optional, Tuple, Union
from PIL import Image

from .transform import (
    Transform,
    TransformList,
    BlendTransform,
    CropTransform,
    HFlipTransform,
    NoOpTransform,
    Transform,
    VFlipTransform,
    ExtentTransform,
    ResizeTransform,
    RotationTransform
)


__all__ = [
    "Augmentation",
    "TransformGen",
    "apply_transform_gens",
    "AugInput",
    "StandardAugInput",
    "apply_augmentations",
    "RandomApply",
    "RandomBrightness",
    "RandomContrast",
    "RandomCrop",
    "RandomExtent",
    "RandomFlip",
    "RandomSaturation",
    "RandomLighting",
    "RandomRotation",
    "Resize",
    "ResizeShortestEdge",
    "RandomCropWithInstance",
    "PadAugmentation",
]

"""
Overview of the augmentation system:

We have a design goal that aims at allowing:
    (1) Arbitrary structures of input data (e.g. list[list[boxes]], dict[str, boxes],
        multiple semantic segmentations for each image, etc) and arbitrary new data types
        (rotated boxes, 3D meshes, densepose, etc)
    (2) A list of augmentation to be applied sequentially

`Augmentation` defines policies to create deterministic transforms from input data.
An augmentation policy may need to access arbitrary input data, so it declares the input
data needed, to be provided by users when calling its `get_transform` method.

`Augmentation` is not able to apply transforms to data: data associated with
one sample may be much more than what `Augmentation` gets. For example, most
augmentation policies only need an image, but the actual input samples can be
much more complicated.

`AugInput` manages all inputs needed by `Augmentation` and implements the logic
to apply a sequence of augmentation. It has to define how the inputs are transformed,
because arguments needed by one `Augmentation` needs to be transformed to become arguments
of the next `Augmentation` in the sequence.

`AugInput` does not need to contain all input data, because most augmentation policies
only need very few fields (e.g., most only need "image"). We provide `StandardAugInput`
that only contains "images", "boxes", "sem_seg", that are enough to create transforms
for most cases. In this way, users keep the responsibility to apply transforms to other
potentially new data types and structures, e.g. keypoints, proposals boxes.

To extend the system, one can do:
1. To add a new augmentation policy that only needs to use standard inputs
   ("image", "boxes", "sem_seg"), writing a subclass of `Augmentation` is sufficient.
2. To use new data types or custom data structures, `StandardAugInput` can still be used as long
   as the new data types or custom data structures are not needed by any augmentation policy.
   The new data types or data structures can be transformed using the
   transforms returned by `AugInput.apply_augmentations`.
3. To add new augmentation policies that need new data types or data structures, in addition to
   implementing new `Augmentation`, a new `AugInput` is needed as well.
"""

def _check_img_dtype(img):
    assert isinstance(img, np.ndarray), "[Augmentation] Needs an numpy array, but got a {}!".format(
        type(img)
    )
    assert not isinstance(img.dtype, np.integer) or (
        img.dtype == np.uint8
    ), "[Augmentation] Got image of type {}, use uint8 or floating points instead!".format(
        img.dtype
    )
    assert img.ndim in [2, 3], img.ndim


class Augmentation(metaclass=ABCMeta):
    """
    Augmentation defines policies/strategies to generate :class:`Transform` from data.
    It is often used for pre-processing of input data. A policy typically contains
    randomness, but it can also choose to deterministically generate a :class:`Transform`.

    A "policy" that generates a :class:`Transform` may, in the most general case,
    need arbitrary information from input data in order to determine what transforms
    to apply. Therefore, each :class:`Augmentation` instance defines the arguments
    needed by its :meth:`get_transform` method with the :attr:`input_args` attribute.
    When called with the positional arguments defined by the :attr:`input_args`,
    the :meth:`get_transform` method executes the policy.

    Examples:
    ::
        # if a policy needs to know both image and semantic segmentation
        assert aug.input_args == ("image", "sem_seg")
        tfm: Transform = aug.get_transform(image, sem_seg)
        new_image = tfm.apply_image(image)

    To implement a custom :class:`Augmentation`, define its :attr:`input_args` and
    implement :meth:`get_transform`.

    Note that :class:`Augmentation` defines the policies to create a :class:`Transform`,
    but not how to apply the actual transform to those data.
    """

    input_args: Tuple[str] = ("image",)
    """
    Attribute of class instances that defines the argument(s) needed by
    :meth:`get_transform`. Default to only "image", because most policies only
    require knowing the image in order to determine the transform.

    Users can freely define arbitrary new args and their types in custom
    :class:`Augmentation`. In detectron2 we use the following convention:

    * image: (H,W) or (H,W,C) ndarray of type uint8 in range [0, 255], or
      floating point in range [0, 1] or [0, 255].
    * boxes: (N,4) ndarray of float32. It represents the instance bounding boxes
      of N instances. Each is in XYXY format in unit of absolute coordinates.
    * sem_seg: (H,W) ndarray of type uint8. Each element is an integer label of pixel.

    We do not specify convention for other types and do not include builtin
    :class:`Augmentation` that uses other types in detectron2.
    """

    def _init(self, params=None):
        if params:
            for k, v in params.items():
                if k != "self" and not k.startswith("_"):
                    setattr(self, k, v)

    # NOTE: in the future, can allow it to return list[Augmentation],
    # to delegate augmentation to others
    @abstractmethod
    def get_transform(self, *args) -> Transform:
        """
        Execute the policy to use input data to create transform(s).

        Args:
            arguments must follow what's defined in :attr:`input_args`.

        Returns:
            Return a :class:`Transform` instance, which is the transform to apply to inputs.
        """
        pass

    def _rand_range(self, low=1.0, high=None, size=None):
        """
        Uniform float random number between low and high.
        """
        if high is None:
            low, high = 0, low
        if size is None:
            size = []
        return np.random.uniform(low, high, size)

    def __repr__(self):
        """
        Produce something like:
        "MyAugmentation(field1={self.field1}, field2={self.field2})"
        """
        try:
            sig = inspect.signature(self.__init__)
            classname = type(self).__name__
            argstr = []
            for name, param in sig.parameters.items():
                assert (
                    param.kind != param.VAR_POSITIONAL and param.kind != param.VAR_KEYWORD
                ), "The default __repr__ doesn't support *args or **kwargs"
                assert hasattr(self, name), (
                    "Attribute {} not found! "
                    "Default __repr__ only works if attributes match the constructor.".format(name)
                )
                attr = getattr(self, name)
                default = param.default
                if default is attr:
                    continue
                argstr.append("{}={}".format(name, pprint.pformat(attr)))
            return "{}({})".format(classname, ", ".join(argstr))
        except AssertionError:
            return super().__repr__()

    __str__ = __repr__


TransformGen = Augmentation
"""
Alias for Augmentation, since it is something that generates :class:`Transform`s
"""


class AugInput:
    """
    A base class for anything on which a list of :class:`Augmentation` can be applied.
    This class provides input arguments for :class:`Augmentation` to use, and defines how
    to apply transforms to these data.

    An instance of this class must satisfy the following:

    * :class:`Augmentation` declares some data it needs as arguments. A :class:`AugInput`
      must provide access to these data in the form of attribute access (``getattr``).
      For example, if a :class:`Augmentation` to be applied needs "image" and "sem_seg"
      arguments, this class must have the attribute "image" and "sem_seg" whose content
      is as required by the :class:`Augmentation`s.
    * This class must have a :meth:`transform(tfm: Transform) -> None` method which
      in-place transforms all attributes stored in the class.
    """

    def transform(self, tfm: Transform) -> None:
        raise NotImplementedError

    def apply_augmentations(
        self, augmentations: List[Union[Augmentation, Transform]]
    ) -> TransformList:
        """
        Apply a list of Transform/Augmentation in-place and returned the applied transform.
        Attributes of this class will be modified.

        Returns:
            TransformList:
                returns transformed inputs and the list of transforms applied.
                The TransformList can then be applied to other data associated with the inputs.
        """
        tfms = []
        for aug in augmentations:
            if isinstance(aug, Augmentation):
                args = []
                for f in aug.input_args:
                    try:
                        args.append(getattr(self, f))
                    except AttributeError:
                        raise AttributeError(
                            f"Augmentation {aug} needs '{f}', which is not an attribute of {self}!"
                        )

                tfm = aug.get_transform(*args)
                assert isinstance(tfm, Transform), (
                    f"{type(aug)}.get_transform must return an instance of Transform! "
                    "Got {type(tfm)} instead."
                )
            else:
                tfm = aug
            self.transform(tfm)
            tfms.append(tfm)
        return TransformList(tfms)


class StandardAugInput(AugInput):
    """
    A standard implementation of :class:`AugInput` for the majority of use cases.
    This class provides the following standard attributes that are common to use by
    Augmentation (augmentation policies). These are chosen because most
    :class:`Augmentation` won't need anything more to define a augmentation policy.
    After applying augmentations to these special attributes, the returned transforms
    can then be used to transform other data structures that users have.

    Attributes:
        image (ndarray): image in HW or HWC format. The meaning of C is up to users
        boxes (ndarray or None): Nx4 boxes in XYXY_ABS mode
        sem_seg (ndarray or None): HxW semantic segmentation mask

    Examples:
    ::
        input = StandardAugInput(image, boxes=boxes)
        tfms = input.apply_augmentations(list_of_augmentations)
        transformed_image = input.image
        transformed_boxes = input.boxes
        transformed_other_data = tfms.apply_other(other_data)

    An extended project that works with new data types may require augmentation
    policies that need more inputs. An algorithm may need to transform inputs
    in a way different from the standard approach defined in this class. In those
    situations, users can implement new subclasses of :class:`AugInput` with differnt
    attributes and the :meth:`transform` method.
    """

    def __init__(
        self,
        image: np.ndarray,
        *,
        boxes: Optional[np.ndarray] = None,
        sem_seg: Optional[np.ndarray] = None,
    ):
        """
        Args:
            image: (H,W) or (H,W,C) ndarray of type uint8 in range [0, 255], or
                floating point in range [0, 1] or [0, 255].
            boxes: (N,4) ndarray of float32. It represents the instance bounding boxes
                of N instances. Each is in XYXY format in unit of absolute coordinates.
            sem_seg: (H,W) ndarray of type uint8. Each element is an integer label of pixel.
        """
        _check_img_dtype(image)
        self.image = image
        self.boxes = boxes
        self.sem_seg = sem_seg

    def transform(self, tfm: Transform) -> None:
        """
        In-place transform all attributes of this class.
        """
        self.image = tfm.apply_image(self.image)
        if self.boxes is not None:
            self.boxes = tfm.apply_box(self.boxes)
        if self.sem_seg is not None:
            self.sem_seg = tfm.apply_segmentation(self.sem_seg)


def apply_augmentations(augmentations: List[Union[Transform, Augmentation]], inputs):
    """
    Use :meth:`AugInput.apply_augmentations` instead.
    """
    if isinstance(inputs, np.ndarray):
        # handle the common case of image-only Augmentation, also for backward compatibility
        image_only = True
        inputs = StandardAugInput(inputs)
    else:
        image_only = False
    tfms = inputs.apply_augmentations(augmentations)
    return inputs.image if image_only else inputs, tfms


apply_transform_gens = apply_augmentations
"""
Alias for backward-compatibility.
"""

class RandomApply(Augmentation):
    """
    Randomly apply the wrapper transformation with a given probability.
    """

    def __init__(self, transform, prob=0.5):
        """
        Args:
            transform (Transform, Augmentation): the transform to be wrapped
                by the `RandomApply`. The `transform` can either be a
                `Transform` or `Augmentation` instance.
            prob (float): probability between 0.0 and 1.0 that
                the wrapper transformation is applied
        """
        super().__init__()
        assert isinstance(transform, (Transform, Augmentation)), (
            f"The given transform must either be a Transform or Augmentation instance. "
            f"Not {type(transform)}"
        )
        assert 0.0 <= prob <= 1.0, f"Probablity must be between 0.0 and 1.0 (given: {prob})"
        self.prob = prob
        self.transform = transform
        if isinstance(transform, Augmentation):
            self.input_args = transform.input_args

    def get_transform(self, img):
        do = self._rand_range() < self.prob
        if do:
            if isinstance(self.transform, Augmentation):
                return self.transform.get_transform(img)
            else:
                return self.transform
        else:
            return NoOpTransform()


class RandomFlip(Augmentation):
    """
    Flip the image horizontally or vertically with the given probability.
    """

    def __init__(self, prob=0.5, *, horizontal=True, vertical=False):
        """
        Args:
            prob (float): probability of flip.
            horizontal (boolean): whether to apply horizontal flipping
            vertical (boolean): whether to apply vertical flipping
        """
        super().__init__()

        if horizontal and vertical:
            raise ValueError("Cannot do both horiz and vert. Please use two Flip instead.")
        if not horizontal and not vertical:
            raise ValueError("At least one of horiz or vert has to be True!")
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        do = self._rand_range() < self.prob
        if do:
            if self.horizontal:
                return HFlipTransform(w)
            elif self.vertical:
                return VFlipTransform(h)
        else:
            return NoOpTransform()


class Resize(Augmentation):
    """ Resize image to a fixed target size"""

    def __init__(self, shape, interp=Image.BILINEAR):
        """
        Args:
            shape: (h, w) tuple or a int
            interp: PIL interpolation method
        """
        if isinstance(shape, int):
            shape = (shape, shape)
        shape = tuple(shape)
        self._init(locals())

    def get_transform(self, img):
        return ResizeTransform(
            img.shape[0], img.shape[1], self.shape[0], self.shape[1], self.interp
        )


class ResizeShortestEdge(Augmentation):
    """
    Scale the shorter edge to the given size, with a limit of `max_size` on the longer edge.
    If `max_size` is reached, then downscale so that the longer edge does not exceed max_size.
    """

    def __init__(
        self, short_edge_length, max_size=sys.maxsize, sample_style="range", interp=Image.BILINEAR
    ):
        """
        Args:
            short_edge_length (list[int]): If ``sample_style=="range"``,
                a [min, max] interval from which to sample the shortest edge length.
                If ``sample_style=="choice"``, a list of shortest edge lengths to sample from.
            max_size (int): maximum allowed longest edge length.
            sample_style (str): either "range" or "choice".
        """
        super().__init__()
        assert sample_style in ["range", "choice"], sample_style

        self.is_range = sample_style == "range"
        if isinstance(short_edge_length, int):
            short_edge_length = (short_edge_length, short_edge_length)
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        if self.is_range:
            size = np.random.randint(self.short_edge_length[0], self.short_edge_length[1] + 1)
        else:
            size = np.random.choice(self.short_edge_length)
        if size == 0:
            return NoOpTransform()

        scale = size * 1.0 / min(h, w)
        if h < w:
            newh, neww = size, scale * w
        else:
            newh, neww = scale * h, size
        if max(newh, neww) > self.max_size:
            scale = self.max_size * 1.0 / max(newh, neww)
            newh = newh * scale
            neww = neww * scale
        neww = int(neww + 0.5)
        newh = int(newh + 0.5)
        return ResizeTransform(h, w, newh, neww, self.interp)


class RandomRotation(Augmentation):
    """
    This method returns a copy of this image, rotated the given
    number of degrees counter clockwise around the given center.
    """

    def __init__(self, angle, expand=True, center=None, sample_style="range", interp=None):
        """
        Args:
            angle (list[float]): If ``sample_style=="range"``,
                a [min, max] interval from which to sample the angle (in degrees).
                If ``sample_style=="choice"``, a list of angles to sample from
            expand (bool): choose if the image should be resized to fit the whole
                rotated image (default), or simply cropped
            center (list[[float, float]]):  If ``sample_style=="range"``,
                a [[minx, miny], [maxx, maxy]] relative interval from which to sample the center,
                [0, 0] being the top left of the image and [1, 1] the bottom right.
                If ``sample_style=="choice"``, a list of centers to sample from
                Default: None, which means that the center of rotation is the center of the image
                center has no effect if expand=True because it only affects shifting
        """
        super().__init__()
        assert sample_style in ["range", "choice"], sample_style
        self.is_range = sample_style == "range"
        if isinstance(angle, (float, int)):
            angle = (angle, angle)
        if center is not None and isinstance(center[0], (float, int)):
            center = (center, center)
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        center = None
        if self.is_range:
            angle = np.random.uniform(self.angle[0], self.angle[1])
            if self.center is not None:
                center = (
                    np.random.uniform(self.center[0][0], self.center[1][0]),
                    np.random.uniform(self.center[0][1], self.center[1][1]),
                )
        else:
            angle = np.random.choice(self.angle)
            if self.center is not None:
                center = np.random.choice(self.center)

        if center is not None:
            center = (w * center[0], h * center[1])  # Convert to absolute coordinates

        if angle % 360 == 0:
            return NoOpTransform()

        return RotationTransform(h, w, angle, expand=self.expand, center=center, interp=self.interp)


class RandomCrop(Augmentation):
    """
    Randomly crop a subimage out of an image.
    """

    def __init__(self, crop_type: str, crop_size):
        """
        Args:
            crop_type (str): one of "relative_range", "relative", "absolute", "absolute_range".
                See `config/defaults.py` for explanation.
            crop_size (tuple[float]): the relative ratio or absolute pixels of
                height and width
        """
        super().__init__()
        assert crop_type in ["relative_range", "relative", "absolute", "absolute_range"]
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        croph, cropw = self.get_crop_size((h, w))
        assert h >= croph and w >= cropw, "Shape computation in {} has bugs.".format(self)
        h0 = np.random.randint(h - croph + 1)
        w0 = np.random.randint(w - cropw + 1)
        return CropTransform(w0, h0, cropw, croph)

    def get_crop_size(self, image_size):
        """
        Args:
            image_size (tuple): height, width

        Returns:
            crop_size (tuple): height, width in absolute pixels
        """
        h, w = image_size
        if self.crop_type == "relative":
            ch, cw = self.crop_size
            return int(h * ch + 0.5), int(w * cw + 0.5)
        elif self.crop_type == "relative_range":
            crop_size = np.asarray(self.crop_size, dtype=np.float32)
            ch, cw = crop_size + np.random.rand(2) * (1 - crop_size)
            return int(h * ch + 0.5), int(w * cw + 0.5)
        elif self.crop_type == "absolute":
            return (min(self.crop_size[0], h), min(self.crop_size[1], w))
        elif self.crop_type == "absolute_range":
            assert self.crop_size[0] <= self.crop_size[1]
            ch = np.random.randint(min(h, self.crop_size[0]), min(h, self.crop_size[1]) + 1)
            cw = np.random.randint(min(w, self.crop_size[0]), min(w, self.crop_size[1]) + 1)
            return ch, cw
        else:
            NotImplementedError("Unknown crop type {}".format(self.crop_type))


class RandomExtent(Augmentation):
    """
    Outputs an image by cropping a random "subrect" of the source image.

    The subrect can be parameterized to include pixels outside the source image,
    in which case they will be set to zeros (i.e. black). The size of the output
    image will vary with the size of the random subrect.
    """

    def __init__(self, scale_range, shift_range):
        """
        Args:
            output_size (h, w): Dimensions of output image
            scale_range (l, h): Range of input-to-output size scaling factor
            shift_range (x, y): Range of shifts of the cropped subrect. The rect
                is shifted by [w / 2 * Uniform(-x, x), h / 2 * Uniform(-y, y)],
                where (w, h) is the (width, height) of the input image. Set each
                component to zero to crop at the image's center.
        """
        super().__init__()
        self._init(locals())

    def get_transform(self, img):
        img_h, img_w = img.shape[:2]

        # Initialize src_rect to fit the input image.
        src_rect = np.array([-0.5 * img_w, -0.5 * img_h, 0.5 * img_w, 0.5 * img_h])

        # Apply a random scaling to the src_rect.
        src_rect *= np.random.uniform(self.scale_range[0], self.scale_range[1])

        # Apply a random shift to the coordinates origin.
        src_rect[0::2] += self.shift_range[0] * img_w * (np.random.rand() - 0.5)
        src_rect[1::2] += self.shift_range[1] * img_h * (np.random.rand() - 0.5)

        # Map src_rect coordinates into image coordinates (center at corner).
        src_rect[0::2] += 0.5 * img_w
        src_rect[1::2] += 0.5 * img_h

        return ExtentTransform(
            src_rect=(src_rect[0], src_rect[1], src_rect[2], src_rect[3]),
            output_size=(int(src_rect[3] - src_rect[1]), int(src_rect[2] - src_rect[0])),
        )


class RandomContrast(Augmentation):
    """
    Randomly transforms image contrast.

    Contrast intensity is uniformly sampled in (intensity_min, intensity_max).
    - intensity < 1 will reduce contrast
    - intensity = 1 will preserve the input image
    - intensity > 1 will increase contrast

    See: https://pillow.readthedocs.io/en/3.0.x/reference/ImageEnhance.html
    """

    def __init__(self, intensity_min, intensity_max):
        """
        Args:
            intensity_min (float): Minimum augmentation
            intensity_max (float): Maximum augmentation
        """
        super().__init__()
        self._init(locals())

    def get_transform(self, img):
        w = np.random.uniform(self.intensity_min, self.intensity_max)
        return BlendTransform(src_image=img.mean(), src_weight=1 - w, dst_weight=w)


class RandomBrightness(Augmentation):
    """
    Randomly transforms image brightness.

    Brightness intensity is uniformly sampled in (intensity_min, intensity_max).
    - intensity < 1 will reduce brightness
    - intensity = 1 will preserve the input image
    - intensity > 1 will increase brightness

    See: https://pillow.readthedocs.io/en/3.0.x/reference/ImageEnhance.html
    """

    def __init__(self, intensity_min, intensity_max):
        """
        Args:
            intensity_min (float): Minimum augmentation
            intensity_max (float): Maximum augmentation
        """
        super().__init__()
        self._init(locals())

    def get_transform(self, img):
        w = np.random.uniform(self.intensity_min, self.intensity_max)
        return BlendTransform(src_image=0, src_weight=1 - w, dst_weight=w)


class RandomSaturation(Augmentation):
    """
    Randomly transforms saturation of an RGB image.
    Input images are assumed to have 'RGB' channel order.

    Saturation intensity is uniformly sampled in (intensity_min, intensity_max).
    - intensity < 1 will reduce saturation (make the image more grayscale)
    - intensity = 1 will preserve the input image
    - intensity > 1 will increase saturation

    See: https://pillow.readthedocs.io/en/3.0.x/reference/ImageEnhance.html
    """

    def __init__(self, intensity_min, intensity_max):
        """
        Args:
            intensity_min (float): Minimum augmentation (1 preserves input).
            intensity_max (float): Maximum augmentation (1 preserves input).
        """
        super().__init__()
        self._init(locals())

    def get_transform(self, img):
        assert img.shape[-1] == 3, "RandomSaturation only works on RGB images"
        w = np.random.uniform(self.intensity_min, self.intensity_max)
        grayscale = img.dot([0.299, 0.587, 0.114])[:, :, np.newaxis]
        return BlendTransform(src_image=grayscale, src_weight=1 - w, dst_weight=w)


class RandomLighting(Augmentation):
    """
    The "lighting" augmentation described in AlexNet, using fixed PCA over ImageNet.
    Input images are assumed to have 'RGB' channel order.

    The degree of color jittering is randomly sampled via a normal distribution,
    with standard deviation given by the scale parameter.
    """

    def __init__(self, scale):
        """
        Args:
            scale (float): Standard deviation of principal component weighting.
        """
        super().__init__()
        self._init(locals())
        self.eigen_vecs = np.array(
            [[-0.5675, 0.7192, 0.4009], [-0.5808, -0.0045, -0.8140], [-0.5836, -0.6948, 0.4203]]
        )
        self.eigen_vals = np.array([0.2175, 0.0188, 0.0045])

    def get_transform(self, img):
        assert img.shape[-1] == 3, "RandomLighting only works on RGB images"
        weights = np.random.normal(scale=self.scale, size=3)
        return BlendTransform(
            src_image=self.eigen_vecs.dot(weights * self.eigen_vals), src_weight=1.0, dst_weight=1.0
        )

def _gen_crop_transform_with_instance(crop_size, image_size, instances, crop_box=True):
    """
    Generate a CropTransform so that the cropping region contains
    the center of the given instance.

    Args:
        crop_size (tuple): h, w in pixels
        image_size (tuple): h, w
        instance (dict): an annotation dict of one instance, in Detectron2's
            dataset format.
    """
    bbox = random.choice(instances)
    crop_size = np.asarray(crop_size, dtype=np.int32)
    center_yx = (bbox[1] + bbox[3]) * 0.5, (bbox[0] + bbox[2]) * 0.5
    assert (
        image_size[0] >= center_yx[0] and image_size[1] >= center_yx[1]
    ), "The annotation bounding box is outside of the image!"
    assert (
        image_size[0] >= crop_size[0] and image_size[1] >= crop_size[1]
    ), "Crop size is larger than image size!"

    min_yx = np.maximum(np.floor(center_yx).astype(np.int32) - crop_size, 0)
    max_yx = np.maximum(np.asarray(image_size, dtype=np.int32) - crop_size, 0)
    max_yx = np.minimum(max_yx, np.ceil(center_yx).astype(np.int32))

    y0 = np.random.randint(min_yx[0], max_yx[0] + 1)
    x0 = np.random.randint(min_yx[1], max_yx[1] + 1)

    # if some instance is cropped extend the box
    if not crop_box:
        num_modifications = 0
        modified = True

        # convert crop_size to float
        crop_size = crop_size.astype(np.float32)
        while modified:
            modified, x0, y0, crop_size = adjust_crop(x0, y0, crop_size, instances)
            num_modifications += 1
            if num_modifications > 100:
                raise ValueError(
                    "Cannot finished cropping adjustment within 100 tries (#instances {}).".format(
                        len(instances)
                    )
                )
                return CropTransform(0, 0, image_size[1], image_size[0])

    return CropTransform(*map(int, (x0, y0, crop_size[1], crop_size[0])))


def adjust_crop(x0, y0, crop_size, instances, eps=1e-3):
    modified = False

    x1 = x0 + crop_size[1]
    y1 = y0 + crop_size[0]

    for bbox in instances:

        if bbox[0] < x0 - eps and bbox[2] > x0 + eps:
            crop_size[1] += x0 - bbox[0]
            x0 = bbox[0]
            modified = True

        if bbox[0] < x1 - eps and bbox[2] > x1 + eps:
            crop_size[1] += bbox[2] - x1
            x1 = bbox[2]
            modified = True

        if bbox[1] < y0 - eps and bbox[3] > y0 + eps:
            crop_size[0] += y0 - bbox[1]
            y0 = bbox[1]
            modified = True

        if bbox[1] < y1 - eps and bbox[3] > y1 + eps:
            crop_size[0] += bbox[3] - y1
            y1 = bbox[3]
            modified = True

    return modified, x0, y0, crop_size


class RandomCropWithInstance(RandomCrop):
    """ Instance-aware cropping.
    """

    def __init__(self, crop_type, crop_size, crop_instance=True):
        """
        Args:
            crop_instance (bool): if False, extend cropping boxes to avoid cropping instances
        """
        super().__init__(crop_type, crop_size)
        self.crop_instance = crop_instance
        self.input_args = ("image", "boxes")

    def get_transform(self, img, boxes):
        image_size = img.shape[:2]
        crop_size = self.get_crop_size(image_size)
        return _gen_crop_transform_with_instance(
            crop_size, image_size, boxes, crop_box=self.crop_instance
        )


class PadAugmentation(Augmentation):
    def __init__(self, crop_size):
        """
        Args:
            crop_instance (bool): if False, extend cropping boxes to avoid cropping instances
        """
        super().__init__()
        self.crop_size = crop_size

    def get_crop_size(self, image_size):
        h, w = image_size
        return (min(self.crop_size[0], h), min(self.crop_size[1], w))

    def get_transform(self, img):
        image_size = img.shape[:2]
        image_size = self.get_crop_size(image_size)
        return _PadTransform(image_size[0], image_size[1], self.crop_size[1], self.crop_size[0])


class _PadTransform(Transform):
    def __init__(self, h: int, w: int, crop_h: int, crop_w: int):
        super().__init__()
        self._set_attributes(locals())

    def apply_image(self, img: np.ndarray) -> np.ndarray:

        h, w = img.shape[:2]
        assert (
                self.h == h and self.w == w
        ), "Input size mismatch h w {}:{} -> {}:{}".format(self.h, self.w, h, w)

        padding = ((0, self.crop_h - h), (0, self.crop_w - w), (0, 0))
        img = np.pad(img, pad_width=padding)

        return img

    def apply_coords(self, coords: np.ndarray) -> np.ndarray:
        return coords



"""A custom nms module for tvm with static shape"""
# pylint: disable=unused-argument
from tvm import relay
from tvm.relay.frontend.pytorch import _op, AttrCvt, get_relay_op

def nms(inputs, input_types):
    """A static NMS hack for torchvision.nms, it requires the nms to be the last layer of the network"""
    boxes = inputs[0]
    scores = inputs[1]
    iou_threshold = inputs[2]
    score_index = 0
    coord_start = 1
    # PyTorch NMS doesn't have parameter top_k and max_output_size
    top_k = max_out_size = -1

    # TVM NMS assumes score > 0
    scores = scores - _op.min(scores) + _op.const(1.0)

    # a tricky to preserve the original index of bounding boxes, as the nms op will re-arrange the array
    indices = _op.transform.arange(_op.squeeze(_op.shape_of(scores)), dtype="int32")
    indices = _op.cast(indices, dtype="float32")
    indices = _op.expand_dims(indices, -1, 1)

    # More efficient way to run NMS for torchvision.nms
    scores = AttrCvt(op_name="expand_dims", extras={"axis": -1, "num_newaxis": 1})([scores], {})

    data = _op.concatenate([scores, boxes, indices], -1)
    data = _op.expand_dims(data, 0, 1)
    valid_ret = _op.vision.get_valid_counts(
        data, score_threshold=1.05, id_index=5, score_index=score_index
    )

    nms_ret = get_relay_op("non_max_suppression")(
        data=valid_ret[1],
        valid_count=valid_ret[0],
        indices=valid_ret[2],
        max_output_size=max_out_size,
        iou_threshold=iou_threshold,
        force_suppress=True,
        top_k=top_k,
        coord_start=1,
        score_index=score_index,
        id_index=5,
        return_indices=False,
        invalid_to_bottom=False,
    )

    nms_ret = relay.squeeze(nms_ret, axis=[0])
    return nms_ret



"""
MXNet implementation of tracktor in SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
# pylint: disable=line-too-long,logging-format-interpolation,unused-argument,missing-function-docstring
from __future__ import absolute_import
import logging
import time
import numpy as np

import mxnet as mx
from gluoncv.utils.bbox import bbox_iou
from gluoncv.data import COCODetection
from .utils import timeit, Track
from .motion_estimation import FarneBeckFlowMotionEstimator
from .motion_estimation import DummyMotionEstimator


def nms_fallback(boxes, thresh):
    """
    Perform non-maximal suppression and return the indices
    Parameters
    ----------
    boxes: [[x, y, xmax, ymax, score]]

    Returns kept box indices
    -------

    """
    order = np.argsort(boxes[:, -1])[::-1]

    iou_mat = bbox_iou(boxes[:, :4], boxes[:, :4])

    keep = []

    while len(order) > 0:
        i = order[0]
        keep.append(i)

        IOU = iou_mat[i, order[1:]]

        remaining = np.where(IOU <= thresh)[0]
        order = order[remaining + 1]

    return keep


def gpu_iou(bbox_a_tensor, bbox_b_tensor):
    """

    Parameters
    ----------
    bbox_a_tensor
    bbox_b_tensor

    Returns
    -------

    """
    if bbox_a_tensor.shape[1] < 4 or bbox_b_tensor.shape[1] < 4:
        raise IndexError("Bounding boxes axis 1 must have at least length 4")

    tl = mx.nd.maximum(bbox_a_tensor.reshape((bbox_a_tensor.shape[0], 1, -1))[:, :, :2], bbox_b_tensor[:, :2])
    br = mx.nd.minimum(bbox_a_tensor.reshape((bbox_a_tensor.shape[0], 1, -1))[:, :, 2:4], bbox_b_tensor[:, 2:4])

    valid = mx.nd.prod(tl < br, axis=2)
    area_i = mx.nd.prod(br - tl, axis=2) * valid
    area_a = mx.nd.prod(bbox_a_tensor[:, 2:4] - bbox_a_tensor[:, :2], axis=1)
    area_b = mx.nd.prod(bbox_b_tensor[:, 2:4] - bbox_b_tensor[:, :2], axis=1)
    return area_i / (area_a.expand_dims(1) + area_b.expand_dims(0) - area_i)


class SMOTTracker:
    """
    Implementation of the SMOT tracker
    The steps to use the tracker is:
    0. Set anchors from the SSD
    1. First call tracker.predict(new_frame)
    2. Then get the tracking anchor information
    3. Run the detractor with the tracking anchor information
    4. Run tracker.update(new_detection, track_info).
    """
    # pylint: disable=dangerous-default-value,unnecessary-comprehension

    def __init__(self,
                 motion_model='no',
                 anchor_array=None,
                 use_motion=True,
                 tracking_classes=[],
                 match_top_k=10,
                 track_keep_alive_thresh=0.1,
                 new_track_iou_thresh=0.3,
                 track_nms_thresh=0.5,
                 gpu_id=0,
                 anchor_assignment_method='iou',
                 joint_linking=False,
                 tracktor=None,
                 ):
        """

        Parameters
        ----------
        anchor_array
        use_motion
        match_top_k
        track_keep_alive_thresh
        new_track_iou_thresh
        track_nms_thresh
        gpu_id
        anchor_assignment
        joint_linking
        """
        self.use_motion = use_motion
        self.tracks = []
        self.all_track_id = 0
        self.pending_index = []
        self.conf_score_thresh = 0.1
        self.anchor_array = anchor_array
        self.next_frame_pred_index = []
        self.next_frame_pred_weights = []
        self.next_frame_pred_bbox = []
        self.waiting_update_tracks = []
        self.next_frame_ref_bbox = []
        self.last_frame = None
        self.k = match_top_k
        self.keep_alive = track_keep_alive_thresh
        self.new_track_iou_thresh = new_track_iou_thresh
        self.track_nms_thresh = track_nms_thresh
        self.frame_cache = None
        self.mx_ctx = mx.gpu(gpu_id)
        self.anchor_assignment_method = anchor_assignment_method
        self.joint_linking = joint_linking

        if len(tracking_classes) == 0:
            raise ValueError("Unknown tracking classes, please let us know what object you want to track")
        self.coco_class_set = set(COCODetection.CLASSES)
        self.coco_class2index_dict = {i:name for i, name in enumerate(COCODetection.CLASSES)}
        self.class_set = set(tracking_classes)
        for class_name in tracking_classes:
            if class_name not in self.coco_class_set:
                raise ValueError("Your cunstom class {} is not supported, only COCO classes are currently supported,\
                 the classes are {}".format(class_name, COCODetection.CLASSES))

        if motion_model == 'farneback':
            self.motion_estimator = FarneBeckFlowMotionEstimator()
        elif motion_model == 'no':
            self.motion_estimator = DummyMotionEstimator()
        else:
            raise ValueError("Unknown motion model: {}".format(motion_model))

    def process_frame_sequence(self, frame_iterator, tracktor):
        """
        Parameters
        ----------
        frame_iterator: each step it emits a tuple of (frame_id, frame_data)
        tracktor

        Returns
        -------
            results_iter:  a response iterator with one tuple (frame_id, frame_rst) per frame
        """
        for frame_id, frame in frame_iterator:
            logging.info('Processing Frame ID: {}'.format(frame_id))
            t_iter_start = time.time()

            # STEP 0: Prepare the tracktor with the new frame data
            motion_pred_data = tracktor.prepare_for_frame(frame)

            # STEP 1: Predict the new locations of the tracked bounding boxes in the tracker
            tracking_anchor_indices, tracking_anchor_weights, tracking_classes = self.motion_predict(frame, motion_pred_data)

            # STEP 2: Run the tracktor
            detection_bboxes, detection_anchor_indices, tracking_bboxes, extra_info \
                = tracktor.detect_and_track(frame,
                                            tracking_anchor_indices,
                                            tracking_anchor_weights,
                                            tracking_classes
                                            )
            if self.anchor_array is None:
                self.set_anchor_array(tracktor.anchors())

            # STEP 3: Update the tracker with detector responses
            self.update(detection_bboxes,
                        tracking_bboxes,
                        detection_anchor_indices,
                        tracking_anchor_indices,
                        tracking_anchor_weights,
                        tracking_classes,
                        extra_info)

            # yield the results of this frame
            results = self._produce_frame_result()

            elapsed = time.time() - t_iter_start
            logging.info("Total Tracking Runtime: {:2.4f} msec, {:.01f} FPS".format(
                elapsed * 1000, 1 / elapsed))
            yield frame_id, results

    @timeit
    def set_anchor_array(self, anchor_arracy):
        self.anchor_array = anchor_arracy
        self.anchor_tensor = mx.nd.array(self.anchor_array, ctx=self.mx_ctx, dtype=np.float32)

    @timeit
    def motion_predict(self, new_frame: np.ndarray, motion_pred_data):
        """
        Perform motion prediction and assign the predicted track locations to corresponding anchors for re-detection.
        It will update the following properties:
            next_frame_pred_index: indices of anchors that bear tracking information. Each track will be assigned to
                several anchors. They will vote in the re-detection processs.
            next_frame_pred_weights: weights in the re-detection voting
            next_frame_pred_bbox: motion-predicted locations of the tracked objects
            waiting_update_tracks: active tracks for re-detection
            next_frame_ref_bbox: original locations of the tracked objects

        Parameters
        ----------
        new_frame: BGR frame of this timestep
        motion_pred_data: extra data needed by the motion predictor
        Returns:
            next_frame_pred_index
            next_frame_pred_weights
        -------

        """
        # STEP 1: Find all active tracks
        active_track_boxes = []
        active_track_indices = []
        active_track_anchor_indices = []
        active_track_anchor_weights = []
        active_track_classes = []

        t_active = time.time()
        for track_idx, track in enumerate(self.tracks):
            if track.is_active():
                active_track_boxes.append(track.mean)
                active_track_indices.append(track_idx)
                src_idx, src_weights = track.source
                active_track_anchor_indices.append(src_idx)
                active_track_anchor_weights.append(src_weights)
                active_track_classes.append([track.class_id])
                logging.debug("active track {} with age: {}".format(track.track_id, track.age))
        active_track_boxes = np.array(active_track_boxes)
        active_track_anchor_indices = np.array(active_track_anchor_indices)
        active_track_anchor_weights = np.array(active_track_anchor_weights)
        e_active = time.time() - t_active
        logging.info('find active runtime: {:.05f}'.format(e_active))

        if len(active_track_boxes) > 0:
            # The following steps only happen if we have something to track

            # STEP 2: Warp the boxes according to flow
            predicted_track_boxes = self._motion_prediction(new_frame, active_track_boxes,
                                                            active_track_anchor_indices,
                                                            active_track_anchor_weights,
                                                            motion_pred_data,
                                                            skip=not self.use_motion)


            # STEP 3: Assign the warped boxes to anchor compositions
            tracking_anchor_indices, tracking_anchor_weights, tracking_anchor_validity = self._assign_box_to_anchors(
                predicted_track_boxes, method=self.anchor_assignment_method)

            # remove tracks becoming invalid after motion prediction
            invalid_track_numbers = np.nonzero(1 - tracking_anchor_validity)[0]
            logging.info("{}/{} tracks become invalid after motion prediction".format(len(invalid_track_numbers), len(active_track_boxes)))

            for i_invalid in invalid_track_numbers:
                self.tracks[active_track_indices[i_invalid]].mark_missed()

            # keep the valid tracks for re-detection
            valid_track_numbers = np.nonzero(tracking_anchor_validity)[0]
            self.next_frame_pred_index = tracking_anchor_indices[valid_track_numbers, ...]
            self.next_frame_pred_weights = tracking_anchor_weights[valid_track_numbers, ...]
            self.next_frame_pred_bbox = predicted_track_boxes[valid_track_numbers, ...]
            self.next_frame_pred_class = np.array(active_track_classes)[valid_track_numbers, ...]

            active_track_indices = np.array(active_track_indices)[valid_track_numbers, ...]
            active_track_boxes = active_track_boxes[valid_track_numbers, ...]

        else:
            # skip flow computation if there is no active track
            # just save the frame in cache
            predicted_track_boxes = self._motion_prediction(new_frame,
                                                            active_track_boxes,
                                                            active_track_anchor_indices, active_track_anchor_weights,
                                                            motion_pred_data, skip=True)

            assert len(predicted_track_boxes) == 0

            self.next_frame_pred_index = np.array([])
            self.next_frame_pred_weights = np.array([])
            self.next_frame_pred_bbox = np.array([])
            self.next_frame_pred_class = np.array([])

        self.waiting_update_tracks = active_track_indices
        self.next_frame_ref_bbox = active_track_boxes

        return self.next_frame_pred_index, self.next_frame_pred_weights, self.next_frame_pred_class

    @timeit
    def update(self, new_detections: np.ndarray, tracking_predictions: np.ndarray,
               detection_anchor_indices: np.ndarray,
               tracking_anchor_indices: np.ndarray, tracking_anchor_weights: np.ndarray,
               tracking_classes: np.ndarray,
               extra_info: dict = None):
        """
        Update the tracks according to tracking and detection predictions.
        Parameters
        ----------
        new_detections: Nx5 ndarray
        tracking_predictions: Mx5 ndarray
        extra_info: a dictionary with extra information

        Returns
        -------
        """
        # pylint: disable=too-many-nested-blocks
        t_pose_processing = time.time()

        logging.info("tracking predictions 's shape is {}".format(tracking_predictions.shape))
        logging.debug(tracking_predictions)
        logging.debug(self.waiting_update_tracks)

        detection_landmarks = extra_info['detection_landmarks'] if 'detection_landmarks' in extra_info else None
        tracking_landmarks = extra_info['tracking_landmarks'] if 'tracking_landmarks' in extra_info else None

        for t in self.tracks:
            t.predict()

        # STEP 1: track level NMS
        still_active_track_pred_indices = []
        still_active_track_indices = []

        if len(tracking_predictions) > 0:

            # class wise NMS
            keep_set = set()
            for c in set(tracking_classes.ravel().tolist()):
                class_pick = np.nonzero(tracking_classes == c)[0]
                keep_tracking_pred_nms_indices = nms_fallback(tracking_predictions[class_pick, ...], self.track_nms_thresh)
                for i_keep in keep_tracking_pred_nms_indices:
                    keep_set.add(class_pick[i_keep])

            still_active_track_pred_indices = []
            for i_pred, i_track in enumerate(self.waiting_update_tracks):
                if i_pred in keep_set:
                    self.tracks[i_track].update(tracking_predictions[i_pred, :],
                                                (tracking_anchor_indices[i_pred, :], tracking_anchor_weights[i_pred, :]),
                                                tracking_landmarks[i_pred, :] if tracking_landmarks is not None else None)
                else:
                    # suppressed tracks in the track NMS process will be marked as Missing
                    self.tracks[i_track].mark_missed()

                if self.tracks[i_track].is_active():
                    still_active_track_pred_indices.append(i_pred)
                    still_active_track_indices.append(i_track)

        # STEP 2: Remove New Detection Overlapping with Tracks
        if len(still_active_track_pred_indices) > 0 and len(new_detections) > 0:
            active_tracking_predictions = tracking_predictions[still_active_track_pred_indices, :]
            det_track_max_iou = bbox_iou(new_detections[:, :4], active_tracking_predictions[:, :4])
            same_class = new_detections[:, -1:] == (tracking_classes[still_active_track_pred_indices, :].T)
            # suppress all new detections that have high IOU with active tracks
            affinity = (det_track_max_iou * same_class).max(axis=1)
            keep_detection_indices = np.nonzero(affinity <= self.new_track_iou_thresh)[0]
        else:
            # otherwise simply keep all detections
            keep_detection_indices = list(range(len(new_detections)))
            active_tracking_predictions = np.array([])

        # STEP 3: New Track Initialization
        if len(keep_detection_indices) > 0:

            active_new_detections = new_detections[keep_detection_indices, :]
            # (Optional) STEP 3.a: Perform joint linking of body and head
            if self.joint_linking:
                tracking_classes = np.array(tracking_classes)
                body2face_link, face2body_link = \
                    self._link_face_body(active_new_detections,
                                         extra_info['detection_keypoints'][keep_detection_indices],
                                         active_tracking_predictions,
                                         extra_info['tracking_keypoints'][still_active_track_pred_indices],
                                         tracking_classes[still_active_track_pred_indices]
                                         )
            else:
                body2face_link, face2body_link = None, None

            new_tracks = []
            for idx, i_new_track in enumerate(keep_detection_indices):
                new_track = Track(new_detections[i_new_track, :4], self.all_track_id,
                                  (detection_anchor_indices[i_new_track, :], np.array([1])),
                                  keep_alive_thresh=self.keep_alive, class_id=new_detections[i_new_track, -1],
                                  attributes=detection_landmarks[i_new_track, :] if detection_landmarks is not None else None)
                if self.joint_linking:
                    if new_track.class_id == 0:
                        # new face track
                        if idx in face2body_link[0]:
                            logging.debug(idx, i_new_track, '0')
                            body_idx = face2body_link[0][idx]
                            if idx > body_idx:
                                new_track.link_to(new_tracks[body_idx])
                        elif idx in face2body_link[2]:
                            logging.debug(idx, i_new_track, '1')
                            body_idx = face2body_link[2][idx]
                            new_track.link_to(self.tracks[still_active_track_indices[body_idx]])

                    if new_track.class_id == 1:
                        # new body track
                        if idx in body2face_link[0]:
                            face_idx = body2face_link[0][idx]
                            if idx > face_idx:
                                new_track.link_to(new_tracks[face_idx])
                        elif idx in body2face_link[2]:
                            face_idx = body2face_link[2][idx]
                            new_track.link_to(self.tracks[still_active_track_indices[face_idx]])

                self.all_track_id += 1
                self.tracks.append(new_track)
                new_tracks.append(new_track)

        elapsed_post_processing = time.time() - t_pose_processing
        logging.info("total tracklets to now is {}, post-processing time: {:.05f} sec".format(
            self.all_track_id, elapsed_post_processing))

    @property
    def active_tracks(self):
        for t in self.tracks:
            if t.is_active():
                yield t

    def _motion_prediction(self, new_frame, tracked_boxes,
                           tracked_boxes_anchor_indices, tracked_boxes_anchor_weights,
                           motion_pred_data, skip=False):
        """
        Perform motion estimation of a set bounding boxes.
        Use either optical flow or SOT algorithms to predict the locations of these bounding boxes in the new frame
        Parameters
        ----------
        new_frame
        tracked_boxes
        tracked_boxes_anchor_indices
        tracked_boxes_anchor_weights
        skip

        Returns
        -------

        """
        if self.frame_cache is None:
            # this is the first frame
            self.frame_cache = self.motion_estimator.initialize(new_frame, motion_pred_data)
            predicted_bboxes = tracked_boxes
        else:
            # this is not the first frame
            predicted_bboxes, self.frame_cache = self.motion_estimator.predict_new_locations(
                self.frame_cache, tracked_boxes, new_frame, motion_pred_data,
                tracked_boxes_anchor_indices=tracked_boxes_anchor_indices,
                tracked_boxes_anchor_weights=tracked_boxes_anchor_weights,
                skip=skip)

        return predicted_bboxes

    def _assign_box_to_anchors(self, boxes: np.ndarray, method: str = 'avg', min_anchor_iou: float = 0.1):
        """
        The actual implementation of the assignment step.
        GPU acceleration is used because the number of anchors is huge
        Parameters
        ----------
        boxes: must have >1 boxes
        anchors

        Returns
        -------

        """
        t_start = time.time()

        t_iou = time.time()
        gpu_boxes = mx.nd.array(boxes, ctx=self.mx_ctx)
        anchor_track_iou = gpu_iou(self.anchor_tensor, gpu_boxes)
        elapsed_iou = time.time() - t_iou
        logging.info("iou computation runtime: {:.05f}".format(elapsed_iou))

        # get the top-k closest anchors instead of 1
        if method == 'max':
            tracking_anchor_ious, tracking_anchor_indices = mx.nd.topk(anchor_track_iou, axis=0, k=1,
                                                                       ret_typ='both', dtype='int32')
            tracking_anchor_ious = tracking_anchor_ious.T.asnumpy()
            tracking_anchor_indices = tracking_anchor_indices.T.asnumpy()
            tracking_anchor_weights = np.ones_like(tracking_anchor_indices)
        elif method == 'avg':
            tracking_anchor_ious, tracking_anchor_indices = mx.nd.topk(anchor_track_iou, axis=0, k=self.k,
                                                                       ret_typ='both', dtype='int32')
            tracking_anchor_ious = tracking_anchor_ious.T.asnumpy()
            tracking_anchor_indices = tracking_anchor_indices.T.asnumpy()
            tracking_anchor_weights = np.ones_like(tracking_anchor_indices) / self.k
        elif method == 'iou':
            t_sort = time.time()

            tracking_anchor_ious, tracking_anchor_indices = mx.nd.topk(anchor_track_iou, axis=0, k=self.k,
                                                                       ret_typ='both', dtype='int32')
            tracking_anchor_ious = tracking_anchor_ious.T.asnumpy()
            tracking_anchor_indices = tracking_anchor_indices.T.asnumpy()

            e_sort = time.time() - t_sort
            logging.info('sorting time: {:.05f}'.format(e_sort))
            tracking_anchor_weights = tracking_anchor_ious / tracking_anchor_ious.sum(axis=1)[..., None]
        else:
            raise ValueError("unknown anchor assignment method")

        max_track_anchor_ious = tracking_anchor_ious.max(axis=1)
        tracking_anchor_validity = max_track_anchor_ious >= min_anchor_iou

        elapsed_assign = time.time() - t_start
        logging.info("assigment runtime: {}".format(elapsed_assign))
        return tracking_anchor_indices, tracking_anchor_weights, tracking_anchor_validity

    @timeit
    def _produce_frame_result(self):
        tracked_objects = []
        for track in self.active_tracks:
            box = {
                'left': track.mean[0],
                'top': track.mean[1],
                'width': track.mean[2] - track.mean[0],
                'height': track.mean[3] - track.mean[1]
            }
            tid = track.display_id
            age = track.age
            classId = track.class_id
            obj = {
                'bbox': box,
                'track_id': tid,
                'age': age,
                'class_id':classId,
                'class_name':self.coco_class2index_dict[classId]
            }
            if track.attributes is not None:
                obj['landmarks'] = track.attributes
            if obj["class_name"] in self.class_set:
                tracked_objects.append(obj)
        return tracked_objects



from timm.scheduler import CosineLRScheduler, PlateauLRScheduler,\
                           StepLRScheduler, TanhLRScheduler

def create_scheduler(cfg, optimizer):
    num_epochs = cfg.train.epochs

    if cfg.train.lr_noise is not None:
        lr_noise = cfg.train.lr_noise
        if isinstance(lr_noise, (list, tuple)):
            noise_range = [n * num_epochs for n in lr_noise]
            if len(noise_range) == 1:
                noise_range = noise_range[0]
        else:
            noise_range = lr_noise * num_epochs
    else:
        noise_range = None

    lr_scheduler = None
    if cfg.train.sched == 'cosine':
        lr_scheduler = CosineLRScheduler(
            optimizer,
            t_initial=num_epochs,
            t_mul=getattr(cfg.train, 'lr_cycle_mul', 1.),
            lr_min=cfg.train.min_lr,
            decay_rate=cfg.train.decay_rate,
            warmup_lr_init=cfg.train.warmup_lr,
            warmup_t=cfg.train.warmup_epochs,
            cycle_limit=getattr(cfg.train, 'lr_cycle_limit', 1),
            t_in_epochs=True,
            noise_range_t=noise_range,
            noise_pct=getattr(cfg.train, 'lr_noise_pct', 0.67),
            noise_std=getattr(cfg.train, 'lr_noise_std', 1.),
            noise_seed=getattr(cfg.misc, 'seed', 42),
        )
        num_epochs = lr_scheduler.get_cycle_length() + cfg.train.cooldown_epochs
    elif cfg.train.sched == 'tanh':
        lr_scheduler = TanhLRScheduler(
            optimizer,
            t_initial=num_epochs,
            t_mul=getattr(cfg.train, 'lr_cycle_mul', 1.),
            lr_min=cfg.train.min_lr,
            warmup_lr_init=cfg.train.warmup_lr,
            warmup_t=cfg.train.warmup_epochs,
            cycle_limit=getattr(cfg.train, 'lr_cycle_limit', 1),
            t_in_epochs=True,
            noise_range_t=noise_range,
            noise_pct=getattr(cfg.train, 'lr_noise_pct', 0.67),
            noise_std=getattr(cfg.train, 'lr_noise_std', 1.),
            noise_seed=getattr(cfg.misc, 'seed', 42),
        )
        num_epochs = lr_scheduler.get_cycle_length() + cfg.train.cooldown_epochs
    elif cfg.train.sched == 'step':
        lr_scheduler = StepLRScheduler(
            optimizer,
            decay_t=cfg.train.decay_epochs,
            decay_rate=cfg.train.decay_rate,
            warmup_lr_init=cfg.train.warmup_lr,
            warmup_t=cfg.train.warmup_epochs,
            noise_range_t=noise_range,
            noise_pct=getattr(cfg.train, 'lr_noise_pct', 0.67),
            noise_std=getattr(cfg.train, 'lr_noise_std', 1.),
            noise_seed=getattr(cfg.misc, 'seed', 42),
        )
    elif cfg.train.sched == 'plateau':
        mode = 'min' if 'loss' in getattr(cfg.misc, 'eval_metric', '') else 'max'
        lr_scheduler = PlateauLRScheduler(
            optimizer,
            decay_rate=cfg.train.decay_rate,
            patience_t=cfg.train.patience_epochs,
            lr_min=cfg.train.min_lr,
            mode=mode,
            warmup_lr_init=cfg.train.warmup_lr,
            warmup_t=cfg.train.warmup_epochs,
            cooldown_t=0,
            noise_range_t=noise_range,
            noise_pct=getattr(cfg.train, 'lr_noise_pct', 0.67),
            noise_std=getattr(cfg.train, 'lr_noise_std', 1.),
            noise_seed=getattr(cfg.misc, 'seed', 42),
        )

    return lr_scheduler, num_epochs



import json
from collections import OrderedDict
from pathlib import Path
from typing import List
import re

import numpy as np
import torch
import torch.utils.data as data
from tqdm import tqdm

from gluoncv.torch.utils import coot_utils


def load_h5file(h5_path):
    import h5py
    return h5py.File(h5_path, "r")

class BertTextFeatureLoader:
    def __init__(
            self, dataset_path_dict, ids: List[str], preload=True):
        self.h5_path = dataset_path_dict["language_feats"]
        lens_file = dataset_path_dict["meta_text_len"]
        self.lens = json.load(lens_file.open("rt", encoding="utf8"))
        self.cached_data = None
        if preload:
            h5file = load_h5file(self.h5_path)
            self.cached_data = {}
            for id_ in tqdm(ids, desc="preload text"):
                np_array = h5file[id_]
                shared_array = coot_utils.make_shared_array(np_array)
                self.cached_data[id_] = shared_array
            h5file.close()

    def __getitem__(self, id_):
        lens = self.lens[id_]
        if self.cached_data is None:
            h5file = load_h5file(self.h5_path)
            features = np.array(h5file[id_])
            h5file.close()
            return features, lens
        return self.cached_data[id_], lens


class ActivityNetVideoFeatureLoader:
    def __init__(self, dataset_path: Path, ids: List[str], preload: bool):
        self.dataset_path = Path(dataset_path)
        self.features_path = (dataset_path / "features" /
                              "ICEP_V3_global_pool_skip_8_direct_resize")
        self.cached_data = None
        if preload:
            self.cached_data = {}
            for id_ in tqdm(ids, desc="preload videos"):
                np_array = self.load_from_file(id_)
                shared_array = coot_utils.make_shared_array(np_array)
                self.cached_data[id_] = shared_array

    def __getitem__(self, id_):
        if self.cached_data is None:
            return self.load_from_file(id_)
        else:
            return self.cached_data[id_]

    def load_from_file(self, id_):
        return np.load(str(self.features_path / f"{id_}.npz"))[
            "frame_scores"].squeeze(1).squeeze(2).squeeze(2)


class Youcook2VideoFeatureLoader:
    def __init__(
            self, dataset_path_dict, ids: List[str],
            preload: bool):
        self.h5_path = dataset_path_dict["video_feats"]
        self.cached_data = None
        if preload:
            self.cached_data = {}
            h5file = load_h5file(self.h5_path)
            for id_ in tqdm(ids, desc="preload videos"):
                np_array = h5file[id_]
                shared_array = coot_utils.make_shared_array(np_array)
                self.cached_data[id_] = shared_array

    def __getitem__(self, id_):
        if self.cached_data is None:
            h5file = load_h5file(self.h5_path)
            features = np.array(h5file[id_])
            h5file.close()
            return features
        else:
            return self.cached_data[id_]


class VideoDatasetFeatures(data.Dataset):
    def __init__(
            self, dataset_path_dict,
            split: str, max_frames: int, is_train: bool,
            preload_vid_feat: bool, preload_text_feat: bool,
            frames_noise: float):
        self.frames_noise = frames_noise
        self.split = split
        self.max_frames = max_frames
        self.is_train = is_train
        meta_file = dataset_path_dict["meta_data"]
        
        self.vids_dict = json.load(meta_file.open("rt", encoding="utf8"),
                                   object_pairs_hook=OrderedDict)
        self.ids = [key for key, val in self.vids_dict.items(
        ) if val["split"] == self.split]
        print("init dataset {} split {} length {} ".format(dataset_path_dict["dataset_name"], split, len(self)))

        if dataset_path_dict["dataset_name"]  == "activitynet":
            self.text_data = BertTextFeatureLoader(
                dataset_path, self.ids, dataset_features,
                preload_text_feat)
            self.preproc_par_fn = self.preprocess_bert_paragraph
            self.vid_data = ActivityNetVideoFeatureLoader(
                dataset_path, self.ids, preload_vid_feat)
        elif dataset_path_dict["dataset_name"] == "youcook2":
            self.preproc_par_fn = self.preprocess_bert_paragraph
            self.text_data = BertTextFeatureLoader(
                dataset_path_dict, self.ids, preload_text_feat)
            self.vid_data = Youcook2VideoFeatureLoader(
                dataset_path_dict, self.ids, preload_vid_feat)
        else:
            raise NotImplementedError


    def preprocess_bert_sentence(self, sentence_str: str) -> List[str]:
        if sentence_str[-1] == ".":
            sentence_str = sentence_str[:-1]
        sentence_str = sentence_str.replace(". ", " [SEP] ")
        sentence_str += " [SEP] "
        sentence_str = re.sub(r"\s+", " ", sentence_str).strip()
        words = sentence_str.split(" ")
        return words

    def preprocess_bert_paragraph(self,
            paragraph: List[str]) -> List[List[str]]:
        new_paragraph = []
        for i, sentence in enumerate(paragraph):
            new_sentence = []
            if i == 0:
                new_sentence.append("[CLS]")
            preproc_sentence = self.preprocess_bert_sentence(sentence)
            for word in preproc_sentence:
                new_sentence.append(word.strip())
            new_paragraph.append(new_sentence)
        return new_paragraph


    def get_frames_from_video(
            self, vid_id, indices=None, num_frames=None):
        vid_dict = self.vids_dict[vid_id]
        vid_len = vid_dict["num_frames"]
        if num_frames is not None:
            indices = coot_utils.compute_indices(
                vid_len, num_frames, self.is_train)
        frames = self.vid_data[vid_id][indices]
        return frames

    def get_frames_from_segment(
            self, vid_id, seg_num, num_frames):
        vid_dict = self.vids_dict[vid_id]
        seg = vid_dict["segments"][seg_num]
        start_frame = seg["start_frame"]
        seg_len = seg["num_frames"]
        indices = coot_utils.compute_indices(seg_len, num_frames, self.is_train)
        indices += start_frame
        frames = self.get_frames_from_video(vid_id, indices)
        return frames

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, index):
        vid_id = self.ids[index]
        vid_dict = self.vids_dict[vid_id]
        clip_num = len(vid_dict["segments"])
        sent_num = len(vid_dict["segments"])

        # load video frames
        vid_frames_len = vid_dict["num_frames"]
        if vid_frames_len > self.max_frames:
            vid_frames_len = self.max_frames
        vid_frames = torch.tensor(self.get_frames_from_video(
            vid_id, num_frames=vid_frames_len))
        vid_frames_len = int(vid_frames.shape[0])
        if self.frames_noise != 0:
            vid_frames_noise = coot_utils.truncated_normal_fill(
                vid_frames.shape, std=self.frames_noise)
            vid_frames += vid_frames_noise

        # load segment frames
        clip_frames_list = []
        clip_frames_len_list = []
        for i, seg in enumerate(vid_dict["segments"]):
            c_num_frames = seg["num_frames"]
            if c_num_frames > self.max_frames:
                c_num_frames = self.max_frames
            c_frames = self.get_frames_from_segment(
                vid_id, i, num_frames=c_num_frames)
            c_frames = torch.tensor(c_frames)
            if self.frames_noise != 0:
                clip_frames_noise = coot_utils.truncated_normal_fill(
                    c_frames.shape, std=self.frames_noise)
                c_frames += clip_frames_noise
            clip_frames_list.append(c_frames)
            clip_frames_len_list.append(c_frames.shape[0])

        # load text
        seg_narrations = []
        for seg in vid_dict["segments"]:
            seg_narr = seg["narration"]
            if seg_narr is None:
                seg_narr = "undefined"
                print("WARNING: Undefined text tokens "
                      "(no narration data, is this a test set?)")
            seg_narrations.append(seg_narr)
        list_of_list_of_words = self.preproc_par_fn(seg_narrations)

        # load precomputed text features
        par_cap_vectors, sent_cap_len_list = self.text_data[vid_id]
        par_cap_len = int(par_cap_vectors.shape[0])
        par_cap_vectors = torch.tensor(par_cap_vectors).float()

        # split paragraph features into sentences
        sent_cap_vectors_list = []
        pointer = 0
        for i, sent_cap_len in enumerate(sent_cap_len_list):
            sent_cap_vectors = par_cap_vectors[
                               pointer:pointer + sent_cap_len, :]
            sent_cap_vectors_list.append(sent_cap_vectors)
            pointer += sent_cap_len

        return {
            "vid_id": vid_id,
            "data_words": list_of_list_of_words,
            "vid_frames": vid_frames,
            "vid_frames_len": vid_frames_len,
            "par_cap_vectors": par_cap_vectors,
            "par_cap_len": par_cap_len,
            "clip_num": clip_num,
            "sent_num": sent_num,
            "clip_frames_list": clip_frames_list,
            "clip_frames_len_list": clip_frames_len_list,
            "sent_cap_len_list": sent_cap_len_list,
            "sent_cap_vectors_list": sent_cap_vectors_list
        }

    def collate_fn(self, data_batch):
        def get_data(key):
            return [d[key] for d in data_batch]

        batch_size = len(data_batch)

        # collate video frames
        list_vid_frames = get_data("vid_frames")
        list_vid_frames_len = get_data("vid_frames_len")
        vid_feature_dim = list_vid_frames[0].shape[-1]
        vid_frames_len = torch.tensor(list_vid_frames_len).long()
        vid_frames_max_seq_len = int(vid_frames_len.max().numpy())
        vid_frames = torch.zeros(
            batch_size, vid_frames_max_seq_len, vid_feature_dim).float()
        vid_frames_mask = torch.zeros(batch_size, vid_frames_max_seq_len)
        for batch, (seq_len, item) in enumerate(zip(
                list_vid_frames_len, list_vid_frames)):
            vid_frames[batch, :seq_len] = item
            vid_frames_mask[batch, :seq_len] = 1

        # collate paragraph features
        list_par_cap_len = get_data("par_cap_len")
        list_par_cap_vectors = get_data("par_cap_vectors")
        par_feature_dim = list_par_cap_vectors[0].shape[-1]
        par_cap_len = torch.tensor(list_par_cap_len).long()
        par_cap_max_len = int(par_cap_len.max().numpy())
        par_cap_vectors = torch.zeros(
            batch_size, par_cap_max_len, par_feature_dim).float()
        par_cap_mask = torch.zeros(batch_size, par_cap_max_len)
        for batch, (seq_len, item) in enumerate(
                zip(list_par_cap_len, list_par_cap_vectors)):
            par_cap_vectors[batch, :seq_len, :] = item
            par_cap_mask[batch, :seq_len] = 1

        # collate clip frames
        list_clip_num = get_data("clip_num")
        clip_num = torch.tensor(list_clip_num).long()
        total_clip_num = int(np.sum(list_clip_num))
        list_clip_frames_len_list = get_data("clip_frames_len_list")
        clip_frames_max_len = int(np.max(
            [np.max(len_single) for len_single in list_clip_frames_len_list]))
        clip_frames = torch.zeros((
            total_clip_num, clip_frames_max_len, vid_feature_dim)).float()
        clip_frames_mask = torch.zeros(
            (total_clip_num, clip_frames_max_len))
        list_clip_frames_list = get_data("clip_frames_list")
        clip_frames_len = []
        c_num = 0
        for batch, clip_frames_list in enumerate(list_clip_frames_list):
            for i, clip_frames_item in enumerate(clip_frames_list):
                clip_frames_len_item = int(clip_frames_item.shape[0])
                clip_frames[c_num, :clip_frames_len_item, :] =\
                    clip_frames_item
                clip_frames_mask[c_num, :clip_frames_len_item] = 1
                clip_frames_len.append(clip_frames_len_item)
                c_num += 1
        clip_frames_len = torch.tensor(clip_frames_len).long()

        # collate sentence features
        list_sent_num = get_data("sent_num")
        sent_num = torch.tensor(list_sent_num).long()
        total_sent_num = int(np.sum(list_sent_num))
        list_sent_cap_len_list = get_data("sent_cap_len_list")
        sent_cap_max_len = int(np.max(
            [np.max(len_single) for len_single in list_sent_cap_len_list]))
        sent_cap_len = []
        sent_cap_mask = torch.zeros(
            (total_sent_num, sent_cap_max_len)).long()
        cap_feature_dim = list_par_cap_vectors[0].shape[-1]
        sent_cap_vectors = torch.zeros(
            (total_sent_num, sent_cap_max_len, cap_feature_dim))
        c_num = 0
        for batch, sent_cap_len_list in enumerate(
                list_sent_cap_len_list):
            pointer = 0
            for sent_cap_len_item in sent_cap_len_list:
                sent_cap_vectors[c_num, :sent_cap_len_item] =\
                    par_cap_vectors[
                    batch, pointer:pointer + sent_cap_len_item]
                sent_cap_mask[c_num, :sent_cap_len_item] = 1
                sent_cap_len.append(sent_cap_len_item)
                c_num += 1
                pointer += sent_cap_len_item
        sent_cap_len = torch.tensor(sent_cap_len).long()

        return {
            "vid_frames": vid_frames,
            "vid_frames_mask": vid_frames_mask,
            "vid_frames_len": vid_frames_len,
            "par_cap_vectors": par_cap_vectors,
            "par_cap_mask": par_cap_mask,
            "par_cap_len": par_cap_len,
            "clip_num": clip_num,
            "clip_frames": clip_frames,
            "clip_frames_len": clip_frames_len,
            "clip_frames_mask": clip_frames_mask,
            "sent_num": sent_num,
            "sent_cap_vectors": sent_cap_vectors,
            "sent_cap_mask": sent_cap_mask,
            "sent_cap_len": sent_cap_len,
            "vid_id": get_data("vid_id"),
            "data_words": get_data("data_words")
        }


def create_datasets(
        dataset_path_dict, cfg, preload_vid_feat: bool,
        preload_text_feat: bool, eval=False):
    
    if eval:
        val_set = VideoDatasetFeatures(dataset_path_dict,
        cfg.dataset.val_split, cfg.dataset.max_frames, False, preload_vid_feat,
        preload_text_feat, 0)
        return val_set
    train_set = VideoDatasetFeatures(dataset_path_dict,
        cfg.CONFIG.COOT_DATA.TRAIN_SPLIT, cfg.CONFIG.COOT_DATA.MAX_FRAMES, True,
        preload_vid_feat, preload_text_feat, False)
    val_set = VideoDatasetFeatures(dataset_path_dict,
        cfg.CONFIG.COOT_DATA.VALIDATION_SPLIT, cfg.CONFIG.COOT_DATA.MAX_FRAMES, False, preload_vid_feat,
        preload_text_feat, 0)
    return train_set, val_set


def create_loaders(
        train_set: VideoDatasetFeatures, val_set: VideoDatasetFeatures,
        batch_size: int, num_workers: int, eval=False):
    if eval:
        val_loader = data.DataLoader(
        val_set, batch_size=batch_size, shuffle=False,
        num_workers=num_workers, collate_fn=val_set.collate_fn,
        pin_memory=True)
        return val_loader

    train_loader = data.DataLoader(
        train_set, batch_size=batch_size, shuffle=True,
        num_workers=num_workers, collate_fn=train_set.collate_fn,
        pin_memory=True)
    val_loader = data.DataLoader(
        val_set, batch_size=batch_size, shuffle=False,
        num_workers=num_workers, collate_fn=val_set.collate_fn,
        pin_memory=True)
    return train_loader, val_loader



# pylint: disable=wildcard-import
"""
SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
from __future__ import absolute_import
from .smot_tracker import *
from .tracktors import *



"""Predefined Mask RCNN Model."""
from __future__ import absolute_import

import mxnet as mx
from mxnet.gluon import nn
from mxnet.gluon.contrib.nn import SyncBatchNorm

from ..mask_rcnn import get_mask_rcnn
from ....nn.feature import FPNFeatureExpander

__all__ = ['mask_rcnn_resnet50_v1b_coco',
           'mask_rcnn_fpn_resnet50_v1b_coco',
           'mask_rcnn_resnet101_v1d_coco',
           'mask_rcnn_fpn_resnet101_v1d_coco',
           'mask_rcnn_resnet18_v1b_coco',
           'mask_rcnn_fpn_resnet18_v1b_coco',
           'mask_rcnn_fpn_syncbn_resnet18_v1b_coco',
           'mask_rcnn_fpn_syncbn_mobilenet1_0_coco']


def mask_rcnn_resnet50_v1b_coco(pretrained=False, pretrained_base=True, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_resnet50_v1b_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet50_v1b
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    base_network = resnet50_v1b(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = nn.HybridSequential()
    top_features = nn.HybridSequential()
    for layer in ['conv1', 'bn1', 'relu', 'maxpool', 'layer1', 'layer2', 'layer3']:
        features.add(getattr(base_network, layer))
    for layer in ['layer4']:
        top_features.add(getattr(base_network, layer))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='resnet50_v1b', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        mask_channels=256, rcnn_max_dets=1000,
        short=800, max_size=1333, train_patterns=train_patterns,
        nms_thresh=0.5, nms_topk=-1, post_nms=-1,
        roi_mode='align', roi_size=(14, 14), strides=16, clip=4.42,
        rpn_channel=1024, base_size=16, scales=(2, 4, 8, 16, 32),
        ratios=(0.5, 1, 2), alloc_size=(128, 128), rpn_nms_thresh=0.7,
        rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=6000, rpn_test_post_nms=1000, rpn_min_size=1,
        num_sample=128, pos_iou_thresh=0.5, pos_ratio=0.25,
        **kwargs)


def mask_rcnn_fpn_resnet50_v1b_coco(pretrained=False, pretrained_base=True, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_resnet50_v1b_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet50_v1b
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    base_network = resnet50_v1b(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = FPNFeatureExpander(
        network=base_network,
        outputs=['layers1_relu8_fwd', 'layers2_relu11_fwd', 'layers3_relu17_fwd',
                 'layers4_relu8_fwd'], num_filters=[256, 256, 256, 256], use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=False, pretrained=pretrained_base)
    top_features = None
    box_features = nn.HybridSequential()
    box_features.add(nn.AvgPool2D(pool_size=(3, 3), strides=2, padding=1))  # reduce to 7x7
    for _ in range(2):
        box_features.add(nn.Dense(1024, weight_initializer=mx.init.Normal(0.01)),
                         nn.Activation('relu'))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask', 'P',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='fpn_resnet50_v1b', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        box_features=box_features, mask_channels=256, rcnn_max_dets=1000,
        short=800, max_size=1333, min_stage=2, max_stage=6,
        train_patterns=train_patterns, nms_thresh=0.5, nms_topk=-1,
        post_nms=-1, roi_mode='align', roi_size=(14, 14),
        strides=(4, 8, 16, 32, 64), clip=4.42, rpn_channel=1024, base_size=16,
        scales=(2, 4, 8, 16, 32), ratios=(0.5, 1, 2), alloc_size=(384, 384),
        rpn_nms_thresh=0.7, rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=6000, rpn_test_post_nms=1000, rpn_min_size=1,
        num_sample=512, pos_iou_thresh=0.5, pos_ratio=0.25, target_roi_scale=2,
        num_fcn_convs=4, **kwargs)


def mask_rcnn_resnet101_v1d_coco(pretrained=False, pretrained_base=True, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `Ture`, this has no effect.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_resnet101_v1d_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet101_v1d
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    base_network = resnet101_v1d(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = nn.HybridSequential()
    top_features = nn.HybridSequential()
    for layer in ['conv1', 'bn1', 'relu', 'maxpool', 'layer1', 'layer2', 'layer3']:
        features.add(getattr(base_network, layer))
    for layer in ['layer4']:
        top_features.add(getattr(base_network, layer))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='resnet101_v1d', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        mask_channels=256, rcnn_max_dets=1000,
        short=800, max_size=1333, train_patterns=train_patterns,
        nms_thresh=0.5, nms_topk=-1, post_nms=-1,
        roi_mode='align', roi_size=(14, 14), strides=16, clip=4.42,
        rpn_channel=1024, base_size=16, scales=(2, 4, 8, 16, 32),
        ratios=(0.5, 1, 2), alloc_size=(128, 128), rpn_nms_thresh=0.7,
        rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=6000, rpn_test_post_nms=1000, rpn_min_size=1,
        num_sample=128, pos_iou_thresh=0.5, pos_ratio=0.25,
        **kwargs)


def mask_rcnn_fpn_resnet101_v1d_coco(pretrained=False, pretrained_base=True, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_fpn_resnet101_v1d_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet101_v1d
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    base_network = resnet101_v1d(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = FPNFeatureExpander(
        network=base_network,
        outputs=['layers1_relu8_fwd', 'layers2_relu11_fwd', 'layers3_relu68_fwd',
                 'layers4_relu8_fwd'], num_filters=[256, 256, 256, 256], use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=False, pretrained=pretrained_base)
    top_features = None
    box_features = nn.HybridSequential()
    box_features.add(nn.AvgPool2D(pool_size=(3, 3), strides=2, padding=1))  # reduce to 7x7
    for _ in range(2):
        box_features.add(nn.Dense(1024, weight_initializer=mx.init.Normal(0.01)),
                         nn.Activation('relu'))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask', 'P',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='fpn_resnet101_v1d', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        box_features=box_features, mask_channels=256, rcnn_max_dets=1000,
        short=800, max_size=1333, min_stage=2, max_stage=6,
        train_patterns=train_patterns, nms_thresh=0.5, nms_topk=-1,
        post_nms=-1, roi_mode='align', roi_size=(14, 14),
        strides=(4, 8, 16, 32, 64), clip=4.42, rpn_channel=1024, base_size=16,
        scales=(2, 4, 8, 16, 32), ratios=(0.5, 1, 2), alloc_size=(384, 384),
        rpn_nms_thresh=0.7, rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=6000, rpn_test_post_nms=1000, rpn_min_size=1,
        num_sample=512, pos_iou_thresh=0.5, pos_ratio=0.25, target_roi_scale=2,
        num_fcn_convs=4, **kwargs)


def mask_rcnn_resnet18_v1b_coco(pretrained=False, pretrained_base=True, rcnn_max_dets=1000,
                                rpn_test_pre_nms=6000, rpn_test_post_nms=1000, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    rcnn_max_dets : int, default is 1000
        Number of rois to retain in RCNN.
    rpn_test_pre_nms : int, default is 6000
        Filter top proposals before NMS in testing of RPN.
    rpn_test_post_nms : int, default is 300
        Return top proposal results after NMS in testing of RPN.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_resnet18_v1b_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet18_v1b
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    rcnn_max_dets = rpn_test_post_nms if rcnn_max_dets > rpn_test_post_nms else rcnn_max_dets
    base_network = resnet18_v1b(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = nn.HybridSequential()
    top_features = nn.HybridSequential()
    for layer in ['conv1', 'bn1', 'relu', 'maxpool', 'layer1', 'layer2', 'layer3']:
        features.add(getattr(base_network, layer))
    for layer in ['layer4']:
        top_features.add(getattr(base_network, layer))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='resnet18_v1b', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        mask_channels=256, rcnn_max_dets=rcnn_max_dets,
        short=800, max_size=1333, train_patterns=train_patterns,
        nms_thresh=0.5, nms_topk=-1, post_nms=-1,
        roi_mode='align', roi_size=(14, 14), strides=16, clip=4.42,
        rpn_channel=1024, base_size=16, scales=(2, 4, 8, 16, 32),
        ratios=(0.5, 1, 2), alloc_size=(128, 128), rpn_nms_thresh=0.7,
        rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=rpn_test_pre_nms, rpn_test_post_nms=rpn_test_post_nms,
        rpn_min_size=1, num_sample=256, pos_iou_thresh=0.5, pos_ratio=0.25,
        **kwargs)


def mask_rcnn_fpn_resnet18_v1b_coco(pretrained=False, pretrained_base=True, rcnn_max_dets=1000,
                                    rpn_test_pre_nms=6000, rpn_test_post_nms=1000, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    rcnn_max_dets : int, default is 1000
        Number of rois to retain in RCNN.
    rpn_test_pre_nms : int, default is 6000
        Filter top proposals before NMS in testing of RPN.
    rpn_test_post_nms : int, default is 300
        Return top proposal results after NMS in testing of RPN.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_fpn_resnet18_v1b_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet18_v1b
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    rcnn_max_dets = rpn_test_post_nms if rcnn_max_dets > rpn_test_post_nms else rcnn_max_dets
    base_network = resnet18_v1b(pretrained=pretrained_base, dilated=False, use_global_stats=True)
    features = FPNFeatureExpander(
        network=base_network,
        outputs=['layers1_relu3_fwd', 'layers2_relu3_fwd', 'layers3_relu3_fwd',
                 'layers4_relu3_fwd'], num_filters=[256, 256, 256, 256], use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=False, pretrained=pretrained_base)
    top_features = None
    box_features = nn.HybridSequential()
    for _ in range(2):
        box_features.add(nn.Dense(1024, weight_initializer=mx.init.Normal(0.01)),
                         nn.Activation('relu'))
    train_patterns = '|'.join(['.*dense', '.*rpn', '.*mask', 'P',
                               '.*down(2|3|4)_conv', '.*layers(2|3|4)_conv'])
    return get_mask_rcnn(
        name='fpn_resnet18_v1b', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        box_features=box_features, mask_channels=256, rcnn_max_dets=rcnn_max_dets,
        short=800, max_size=1333, min_stage=2, max_stage=6,
        train_patterns=train_patterns, nms_thresh=0.5, nms_topk=-1,
        post_nms=-1, roi_mode='align', roi_size=(7, 7),
        strides=(4, 8, 16, 32, 64), clip=4.42, rpn_channel=1024, base_size=16,
        scales=(2, 4, 8, 16, 32), ratios=(0.5, 1, 2), alloc_size=(384, 384),
        rpn_nms_thresh=0.7, rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=rpn_test_pre_nms, rpn_test_post_nms=rpn_test_post_nms,
        rpn_min_size=1, num_sample=512, pos_iou_thresh=0.5, pos_ratio=0.25,
        target_roi_scale=2, num_fcn_convs=2, **kwargs)


def mask_rcnn_fpn_syncbn_resnet18_v1b_coco(pretrained=False, pretrained_base=True, num_devices=0,
                                           rcnn_max_dets=1000, rpn_test_pre_nms=6000,
                                           rpn_test_post_nms=1000, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    num_devices : int, default is 0
        Number of devices for sync batch norm layer. if less than 1, use all devices available.
    rcnn_max_dets : int, default is 1000
        Number of rois to retain in RCNN.
    rpn_test_pre_nms : int, default is 6000
        Filter top proposals before NMS in testing of RPN.
    rpn_test_post_nms : int, default is 300
        Return top proposal results after NMS in testing of RPN.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_fpn_syncbn_resnet18_v1b_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.resnetv1b import resnet18_v1b
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    rcnn_max_dets = rpn_test_post_nms if rcnn_max_dets > rpn_test_post_nms else rcnn_max_dets
    gluon_norm_kwargs = {'num_devices': num_devices} if num_devices >= 1 else {}
    sym_norm_kwargs = {'ndev': num_devices} if num_devices >= 1 else {}
    base_network = resnet18_v1b(pretrained=pretrained_base, dilated=False, use_global_stats=False,
                                norm_layer=SyncBatchNorm, norm_kwargs=gluon_norm_kwargs, **kwargs)
    features = FPNFeatureExpander(
        network=base_network,
        outputs=['layers1_relu3_fwd', 'layers2_relu3_fwd', 'layers3_relu3_fwd',
                 'layers4_relu3_fwd'], num_filters=[256, 256, 256, 256], use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=False, pretrained=pretrained_base,
        norm_layer=mx.sym.contrib.SyncBatchNorm, norm_kwargs=sym_norm_kwargs)
    top_features = None
    box_features = nn.HybridSequential()
    box_features.add(nn.Conv2D(256, 3, padding=1),
                     SyncBatchNorm(**gluon_norm_kwargs),
                     nn.Activation('relu'),
                     nn.Dense(1024, weight_initializer=mx.init.Normal(0.01)),
                     nn.Activation('relu'))
    train_patterns = '(?!.*moving)'
    return get_mask_rcnn(
        name='fpn_syncbn_resnet18_v1b', dataset='coco', pretrained=pretrained,
        features=features, top_features=top_features, classes=classes,
        box_features=box_features, mask_channels=256, rcnn_max_dets=rcnn_max_dets,
        short=(640, 800), max_size=1333, min_stage=2, max_stage=6,
        train_patterns=train_patterns, nms_thresh=0.5, nms_topk=-1,
        post_nms=-1, roi_mode='align', roi_size=(7, 7),
        strides=(4, 8, 16, 32, 64), clip=4.42, rpn_channel=1024, base_size=16,
        scales=(2, 4, 8, 16, 32), ratios=(0.5, 1, 2), alloc_size=(384, 384),
        rpn_nms_thresh=0.7, rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=rpn_test_pre_nms, rpn_test_post_nms=rpn_test_post_nms,
        rpn_min_size=1, num_sample=512, pos_iou_thresh=0.5, pos_ratio=0.25,
        target_roi_scale=2, num_fcn_convs=2, norm_layer=SyncBatchNorm,
        norm_kwargs=gluon_norm_kwargs, **kwargs)


def mask_rcnn_fpn_syncbn_mobilenet1_0_coco(pretrained=False, pretrained_base=True, num_devices=0,
                                           rcnn_max_dets=1000, rpn_test_pre_nms=6000,
                                           rpn_test_post_nms=1000, **kwargs):
    r"""Mask RCNN model from the paper
    "He, K., Gkioxari, G., Doll&ar, P., & Girshick, R. (2017). Mask R-CNN"

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    num_devices : int, default is 0
        Number of devices for sync batch norm layer. if less than 1, use all devices available.
    rcnn_max_dets : int, default is 1000
        Number of rois to retain in RCNN.
    rpn_test_pre_nms : int, default is 6000
        Filter top proposals before NMS in testing of RPN.
    rpn_test_post_nms : int, default is 300
        Return top proposal results after NMS in testing of RPN.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = mask_rcnn_fpn_syncbn_mobilenet1_0_coco(pretrained=True)
    >>> print(model)
    """
    from ....model_zoo.mobilenet import mobilenet1_0
    from ....data import COCODetection
    classes = COCODetection.CLASSES
    pretrained_base = False if pretrained else pretrained_base
    rcnn_max_dets = rpn_test_post_nms if rcnn_max_dets > rpn_test_post_nms else rcnn_max_dets
    gluon_norm_kwargs = {'num_devices': num_devices} if num_devices >= 1 else {}
    sym_norm_kwargs = {'ndev': num_devices} if num_devices >= 1 else {}
    base_network = mobilenet1_0(pretrained=pretrained_base, norm_layer=SyncBatchNorm,
                                norm_kwargs=gluon_norm_kwargs, **kwargs)
    features = FPNFeatureExpander(
        network=base_network,
        outputs=['relu6_fwd', 'relu10_fwd', 'relu22_fwd', 'relu26_fwd'],
        num_filters=[256, 256, 256, 256], use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=False, pretrained=pretrained_base,
        norm_layer=mx.sym.contrib.SyncBatchNorm, norm_kwargs=sym_norm_kwargs)
    top_features = None
    box_features = nn.HybridSequential()
    box_features.add(nn.AvgPool2D(pool_size=(3, 3), strides=2, padding=1))  # reduce to 7x7
    box_features.add(nn.Conv2D(256, 3, padding=1),
                     SyncBatchNorm(**gluon_norm_kwargs),
                     nn.Activation('relu'),
                     nn.Dense(1024, weight_initializer=mx.init.Normal(0.01)),
                     nn.Activation('relu'))
    train_patterns = '(?!.*moving)'
    return get_mask_rcnn(
        name='fpn_syncbn_mobilenet1_0', dataset='coco', pretrained=pretrained, features=features,
        top_features=top_features, classes=classes, box_features=box_features, mask_channels=256,
        rcnn_max_dets=rcnn_max_dets, short=(640, 800), max_size=1333, min_stage=2, max_stage=6,
        train_patterns=train_patterns, nms_thresh=0.5, nms_topk=-1, post_nms=-1, roi_mode='align',
        roi_size=(14, 14), strides=(4, 8, 16, 32, 64), clip=4.42, rpn_channel=1024, base_size=16,
        scales=(2, 4, 8, 16, 32), ratios=(0.5, 1, 2), alloc_size=(384, 384),
        rpn_nms_thresh=0.7, rpn_train_pre_nms=12000, rpn_train_post_nms=2000,
        rpn_test_pre_nms=rpn_test_pre_nms, rpn_test_post_nms=rpn_test_post_nms, rpn_min_size=1,
        num_sample=512, pos_iou_thresh=0.5, pos_ratio=0.25, target_roi_scale=2, num_fcn_convs=2,
        norm_layer=SyncBatchNorm, norm_kwargs=gluon_norm_kwargs, **kwargs)



from collections import OrderedDict
import os

import torch

# FIXME: shouldn't need this anymore
def resume_checkpoint(model, checkpoint_path, optimizer=None, loss_scaler=None, logger=None, log_info=False):
    resume_epoch = None
    if os.path.isfile(checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location='cpu')
        if isinstance(checkpoint, dict) and 'state_dict' in checkpoint:
            if log_info:
                logger.info('Restoring model state from checkpoint...')
            new_state_dict = OrderedDict()
            for k, v in checkpoint['state_dict'].items():
                name = k[7:] if k.startswith('module') else k
                new_state_dict[name] = v
            model.load_state_dict(new_state_dict)

            if optimizer is not None and 'optimizer' in checkpoint:
                if log_info:
                    logger.info('Restoring optimizer state from checkpoint...')
                optimizer.load_state_dict(checkpoint['optimizer'])

            if loss_scaler is not None and loss_scaler.state_dict_key in checkpoint:
                if log_info:
                    logger.info('Restoring AMP loss scaler state from checkpoint...')
                loss_scaler.load_state_dict(checkpoint[loss_scaler.state_dict_key])

            if 'epoch' in checkpoint:
                resume_epoch = checkpoint['epoch']
                if 'version' in checkpoint and checkpoint['version'] > 1:
                    resume_epoch += 1  # start at the next epoch, old checkpoints incremented before save

            if log_info:
                logger.info("Loaded checkpoint '{}' (epoch {})".format(checkpoint_path, checkpoint['epoch']))
        else:
            model.load_state_dict(checkpoint)
            if log_info:
                logger.info("Loaded checkpoint '{}'".format(checkpoint_path))
        return resume_epoch
    else:
        logger.error("No checkpoint found at '{}'".format(checkpoint_path))
        raise FileNotFoundError()
